from abc import ABC, abstractmethod
from collections.abc import Callable
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import ClassVar, Literal

from cognite.client.exceptions import CogniteException
from rich.live import Live
from rich.spinner import Spinner
from rich.table import Table

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.utils.cdf import (
    label_aggregate_count,
    label_count,
    metadata_key_counts,
    relationship_aggregate_count,
)

from ._base import ToolkitCommand


class AssetCentricAggregator(ABC):
    _transformation_destination: ClassVar[tuple[str, ...]]

    def __init__(self, client: ToolkitClient) -> None:
        self.client = client

    @property
    @abstractmethod
    def display_name(self) -> str:
        raise NotImplementedError()

    @abstractmethod
    def count(self) -> int:
        raise NotImplementedError

    def transformation_count(self) -> int:
        """Returns the number of transformations associated with the resource."""
        transformation_count = 0
        for destination in self._transformation_destination:
            for chunk in self.client.transformations(chunk_size=1000, destination_type=destination, limit=None):
                transformation_count += len(chunk)
        return transformation_count


class MetadataAggregator(AssetCentricAggregator, ABC):
    def __init__(
        self, client: ToolkitClient, resource_name: Literal["assets", "events", "files", "timeseries", "sequences"]
    ) -> None:
        super().__init__(client)
        self.resource_name = resource_name

    def metadata_key_count(self) -> int:
        return len(metadata_key_counts(self.client, self.resource_name))


class LabelAggregator(MetadataAggregator, ABC):
    def label_count(self) -> int:
        return len(label_count(self.client, self.resource_name))


class AssetAggregator(LabelAggregator):
    _transformation_destination = ("assets", "asset_hierarchy")

    def __init__(self, client: ToolkitClient) -> None:
        super().__init__(client, "assets")

    @property
    def display_name(self) -> str:
        return "Assets"

    def count(self) -> int:
        return self.client.assets.aggregate_count()


class EventAggregator(MetadataAggregator):
    _transformation_destination = ("events",)

    def __init__(self, client: ToolkitClient) -> None:
        super().__init__(client, "events")

    @property
    def display_name(self) -> str:
        return "Events"

    def count(self) -> int:
        return self.client.events.aggregate_count()


class FileAggregator(LabelAggregator):
    _transformation_destination = ("files",)

    def __init__(self, client: ToolkitClient) -> None:
        super().__init__(client, "files")

    @property
    def display_name(self) -> str:
        return "Files"

    def count(self) -> int:
        response = self.client.files.aggregate()
        if response:
            return response[0].count
        else:
            return 0


class TimeSeriesAggregator(MetadataAggregator):
    _transformation_destination = ("timeseries",)

    def __init__(self, client: ToolkitClient) -> None:
        super().__init__(client, "timeseries")

    @property
    def display_name(self) -> str:
        return "TimeSeries"

    def count(self) -> int:
        return self.client.time_series.aggregate_count()


class SequenceAggregator(MetadataAggregator):
    _transformation_destination = ("sequences",)

    def __init__(self, client: ToolkitClient) -> None:
        super().__init__(client, "sequences")

    @property
    def display_name(self) -> str:
        return "Sequences"

    def count(self) -> int:
        return self.client.sequences.aggregate_count()


class RelationshipAggregator(AssetCentricAggregator):
    _transformation_destination = ("relationships",)

    @property
    def display_name(self) -> str:
        return "Relationships"

    def count(self) -> int:
        results = relationship_aggregate_count(self.client)
        return sum(result.count for result in results)


class LabelCountAggregator(AssetCentricAggregator):
    _transformation_destination = ("labels",)

    @property
    def display_name(self) -> str:
        return "Labels"

    def count(self) -> int:
        return label_aggregate_count(self.client)


class ProfileCommand(ToolkitCommand):
    class Columns:
        Resource = "Resource"
        Count = "Count"
        MetadataKeyCount = "Metadata Key Count"
        LabelCount = "Label Count"
        Transformation = "Transformations"

    columns = (
        Columns.Resource,
        Columns.Count,
        Columns.MetadataKeyCount,
        Columns.LabelCount,
        Columns.Transformation,
    )
    spinner_speed = 1.0

    @classmethod
    def asset_centric(
        cls,
        client: ToolkitClient,
        verbose: bool = False,
    ) -> list[dict[str, str]]:
        aggregators: list[AssetCentricAggregator] = [
            AssetAggregator(client),
            EventAggregator(client),
            FileAggregator(client),
            TimeSeriesAggregator(client),
            SequenceAggregator(client),
            RelationshipAggregator(client),
            LabelCountAggregator(client),
        ]
        results, api_calls = cls._create_initial_table(aggregators)
        with Live(cls.create_profile_table(results), refresh_per_second=4) as live:
            with ThreadPoolExecutor(max_workers=8) as executor:
                future_to_cell = {
                    executor.submit(api_calls[(index, col)]): (index, col)
                    for index in range(len(aggregators))
                    for col in cls.columns
                    if (index, col) in api_calls
                }
                for future in as_completed(future_to_cell):
                    index, col = future_to_cell[future]
                    results[index][col] = future.result()
                    live.update(cls.create_profile_table(results))
        return [{col: str(value) for col, value in row.items()} for row in results]

    @classmethod
    def _create_initial_table(
        cls, aggregators: list[AssetCentricAggregator]
    ) -> tuple[list[dict[str, str | Spinner]], dict[tuple[int, str], Callable[[], str]]]:
        rows: list[dict[str, str | Spinner]] = []
        api_calls: dict[tuple[int, str], Callable[[], str]] = {}
        for index, aggregator in enumerate(aggregators):
            row: dict[str, str | Spinner] = {
                cls.Columns.Resource: aggregator.display_name,
                cls.Columns.Count: Spinner("arc", text="loading...", style="bold green", speed=cls.spinner_speed),
            }
            api_calls[(index, cls.Columns.Count)] = cls._call_api(aggregator.count)
            count: str | Spinner = "-"
            if isinstance(aggregator, MetadataAggregator):
                count = Spinner("arc", text="loading...", style="bold green", speed=cls.spinner_speed)
                api_calls[(index, cls.Columns.MetadataKeyCount)] = cls._call_api(aggregator.metadata_key_count)
            row[cls.Columns.MetadataKeyCount] = count

            count = "-"
            if isinstance(aggregator, LabelAggregator):
                count = Spinner("arc", text="loading...", style="bold green", speed=cls.spinner_speed)
                api_calls[(index, cls.Columns.LabelCount)] = cls._call_api(aggregator.label_count)
            row[cls.Columns.LabelCount] = count

            row[cls.Columns.Transformation] = Spinner(
                "arc", text="loading...", style="bold green", speed=cls.spinner_speed
            )
            api_calls[(index, cls.Columns.Transformation)] = cls._call_api(aggregator.transformation_count)

            rows.append(row)
        return rows, api_calls

    @classmethod
    def create_profile_table(cls, rows: list[dict[str, str | Spinner]]) -> Table:
        table = Table(
            title="Asset Centric Profile",
            title_justify="left",
            show_header=True,
            header_style="bold magenta",
        )
        for col in cls.columns:
            table.add_column(col)

        for row in rows:
            table.add_row(*row.values())
        return table

    @staticmethod
    def _call_api(call_fun: Callable[[], int]) -> Callable[[], str]:
        def styled_callable() -> str:
            try:
                value = call_fun()
            except CogniteException as e:
                return type(e).__name__
            else:
                return f"{value:,}"

        return styled_callable



import contextlib

import pytest
from cognite.client.data_classes import DataSet, DataSetWrite
from cognite.client.exceptions import CogniteAPIError, CogniteDuplicatedError

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.data_classes.robotics import (
    DataPostProcessing,
    DataPostProcessingList,
    DataPostProcessingWrite,
    Frame,
    FrameList,
    FrameWrite,
    Location,
    LocationList,
    LocationWrite,
    Map,
    MapList,
    MapWrite,
    Point3D,
    Quaternion,
    Robot,
    RobotCapability,
    RobotCapabilityList,
    RobotCapabilityWrite,
    RobotList,
    RobotWrite,
    Transform,
)
from tests.test_integration.constants import RUN_UNIQUE_ID

DESCRIPTIONS = ["Original Description", "Updated Description"]


@pytest.fixture(scope="session")
def existing_capability(toolkit_client: ToolkitClient) -> RobotCapability:
    capability = RobotCapabilityWrite(
        name="ptz",
        external_id=f"ptz_{RUN_UNIQUE_ID}",
        method="ptz",
        input_schema=INPUT_SCHEMA_CAPABILITY,
        data_handling_schema=DATA_HANDLING_SCHEMA_CAPABILITY,
        description=DESCRIPTIONS[0],
    )
    try:
        retrieved = toolkit_client.robotics.capabilities.retrieve(capability.external_id)
        return retrieved
    except CogniteAPIError:
        created = toolkit_client.robotics.capabilities.create(capability)
        return created


@pytest.fixture(scope="session")
def existing_data_processing(toolkit_client: ToolkitClient) -> RobotCapability:
    data_processing = RobotCapabilityWrite(
        name="Read dial gauge",
        external_id=f"read_dial_gauge_{RUN_UNIQUE_ID}",
        method="read_dial_gauge",
        input_schema=INPUT_SCHEMA_CAPABILITY,
        description=DESCRIPTIONS[0],
    )
    try:
        return toolkit_client.robotics.data_postprocessing.retrieve(data_processing.external_id)
    except CogniteAPIError:
        return toolkit_client.robotics.data_postprocessing.create(data_processing)


@pytest.fixture(scope="session")
def existing_map(toolkit_client: ToolkitClient) -> Map:
    map_ = MapWrite(
        name="Robot navigation map",
        external_id=f"robotMap_{RUN_UNIQUE_ID}",
        map_type="POINTCLOUD",
        description=DESCRIPTIONS[0],
    )
    try:
        return toolkit_client.robotics.maps.retrieve(map_.external_id)
    except CogniteAPIError:
        return toolkit_client.robotics.maps.create(map_)


@pytest.fixture(scope="session")
def existing_location(toolkit_client: ToolkitClient) -> Map:
    location = LocationWrite(
        name="Water treatment plant",
        external_id=f"waterTreatmentPlant1_{RUN_UNIQUE_ID}",
        description=DESCRIPTIONS[0],
    )
    try:
        return toolkit_client.robotics.locations.retrieve(location.external_id)
    except CogniteAPIError:
        return toolkit_client.robotics.locations.create(location)


@pytest.fixture(scope="session")
def root_frame(toolkit_client: ToolkitClient) -> Map:
    root = FrameWrite(
        name="Root coordinate frame",
        external_id="rootCoordinateFrame",
    )
    try:
        return toolkit_client.robotics.frames.retrieve(root.external_id)
    except CogniteAPIError:
        return toolkit_client.robotics.frames.create(root)


FRAME_NAMES = ["Root coordinate frame of a location", "Updated name"]


@pytest.fixture(scope="session")
def existing_frame(toolkit_client: ToolkitClient, root_frame: Frame) -> Map:
    location = FrameWrite(
        name=FRAME_NAMES[0],
        external_id=f"rootCoordinateFrame_{RUN_UNIQUE_ID}",
        transform=Transform(
            parent_frame_external_id=root_frame.external_id,
            translation=Point3D(0, 0, 0),
            orientation=Quaternion(0, 0, 0, 1),
        ),
    )
    try:
        return toolkit_client.robotics.frames.retrieve(location.external_id)
    except CogniteAPIError:
        return toolkit_client.robotics.frames.create(location)


@pytest.fixture(scope="session")
def existing_robots_data_set(toolkit_client: ToolkitClient) -> DataSet:
    data_set = DataSetWrite(
        external_id="ds_robotics_api_tests",
        name="Robotics API Tests",
        description="Data set for testing the Robotics API",
    )
    retrieved = toolkit_client.data_sets.retrieve(external_id=data_set.external_id)
    if retrieved:
        return retrieved
    else:
        return toolkit_client.data_sets.create(data_set)


@pytest.fixture(scope="session")
def persistent_robots_data_set(toolkit_client: ToolkitClient) -> DataSet:
    data_set = DataSetWrite(
        external_id="ds_robotics_api_tests_persistent",
        name="Robotics API Tests Persistent",
        description="Data set for testing the Robotics API with persistent data",
    )
    retrieved = toolkit_client.data_sets.retrieve(external_id=data_set.external_id)
    if retrieved:
        return retrieved
    else:
        return toolkit_client.data_sets.create(data_set)


@pytest.fixture(scope="session")
def existing_robot(
    toolkit_client: ToolkitClient, persistent_robots_data_set: DataSet, existing_capability: RobotCapability
) -> Robot:
    robot = RobotWrite(
        name="wall-e",
        capabilities=[existing_capability.external_id],
        robot_type="DJI_DRONE",
        data_set_id=persistent_robots_data_set.id,
        description=DESCRIPTIONS[0],
    )
    try:
        found = toolkit_client.robotics.robots.list()
        return found.get_robot_by_name(robot.name)
    except (CogniteAPIError, ValueError):
        return toolkit_client.robotics.robots.create(robot)


class TestRobotCapabilityAPI:
    def test_create_retrieve_delete(self, toolkit_client: ToolkitClient) -> None:
        capability = RobotCapabilityWrite(
            name="test_create_retrieve_delete",
            external_id=f"test_create_retrieve_delete_{RUN_UNIQUE_ID}",
            method="ptz",
            input_schema=INPUT_SCHEMA_CAPABILITY,
            data_handling_schema=DATA_HANDLING_SCHEMA_CAPABILITY,
            description="Pan, tilt, zoom camera for visual image capture",
        )
        try:
            with contextlib.suppress(CogniteDuplicatedError):
                created = toolkit_client.robotics.capabilities.create(capability)
                assert isinstance(created, RobotCapability)
                assert created.as_write().dump() == capability.dump()

            retrieved = toolkit_client.robotics.capabilities.retrieve(capability.external_id)

            assert isinstance(retrieved, RobotCapability)
            assert retrieved.as_write().dump() == capability.dump()
        finally:
            toolkit_client.robotics.capabilities.delete(capability.external_id)

        with pytest.raises(CogniteAPIError):
            toolkit_client.robotics.capabilities.retrieve(capability.external_id)

    @pytest.mark.usefixtures("existing_capability")
    def test_list_capabilities(self, toolkit_client: ToolkitClient) -> None:
        capabilities = toolkit_client.robotics.capabilities.list()
        assert isinstance(capabilities, RobotCapabilityList)
        assert len(capabilities) > 0

    @pytest.mark.usefixtures("existing_capability")
    def test_iterate_capabilities(self, toolkit_client: ToolkitClient) -> None:
        for capability in toolkit_client.robotics.capabilities:
            assert isinstance(capability, RobotCapability)
            break
        else:
            pytest.fail("No capabilities found")

    def test_update_capability(self, toolkit_client: ToolkitClient, existing_capability: RobotCapability) -> None:
        update = existing_capability
        update.description = next(desc for desc in DESCRIPTIONS if desc != existing_capability.description)
        updated = toolkit_client.robotics.capabilities.update(update)
        assert updated.description == update.description


@pytest.mark.skip("Robot API seems to fail if you have two robots. This causes every other test run to fail.")
class TestRobotsAPI:
    def test_create_retrieve_delete(
        self, toolkit_client: ToolkitClient, existing_robots_data_set: DataSet, existing_capability: RobotCapability
    ) -> None:
        robot = RobotWrite(
            name="test_robot",
            capabilities=[],
            robot_type="SPOT",
            data_set_id=existing_robots_data_set.id,
            description="test_description",
        )
        retrieved: Robot | None = None
        try:
            with contextlib.suppress(CogniteDuplicatedError):
                created = toolkit_client.robotics.robots.create(robot)
                assert isinstance(created, Robot)
                assert created.as_write().dump() == robot.dump()

            all_retrieved = toolkit_client.robotics.robots.retrieve(robot.data_set_id)
            assert isinstance(all_retrieved, RobotList)
            retrieved = all_retrieved.get_robot_by_name(robot.name)
            assert isinstance(retrieved, Robot)
            assert retrieved.as_write().dump() == robot.dump()
        finally:
            if retrieved:
                toolkit_client.robotics.robots.delete(retrieved.data_set_id)

        with pytest.raises(CogniteAPIError):
            toolkit_client.robotics.robots.retrieve(robot.data_set_id)

    @pytest.mark.usefixtures("existing_robot")
    def test_list_robots(self, toolkit_client: ToolkitClient) -> None:
        robots = toolkit_client.robotics.robots.list()
        assert isinstance(robots, RobotList)
        assert len(robots) > 0

    @pytest.mark.usefixtures("existing_robot")
    def test_iterate_robots(self, toolkit_client: ToolkitClient) -> None:
        for robot in toolkit_client.robotics.robots:
            assert isinstance(robot, Robot)
            break
        else:
            pytest.fail("No robots found")

    def test_update_robot(self, toolkit_client: ToolkitClient, existing_robot: Robot) -> None:
        update = existing_robot.as_write()
        update.description = next(desc for desc in DESCRIPTIONS if desc != existing_robot.description)
        updated = toolkit_client.robotics.robots.update(update)
        assert updated.description == update.description


class TestDataProcessingAPI:
    def test_create_retrieve_delete(self, toolkit_client: ToolkitClient) -> None:
        data_processing = DataPostProcessingWrite(
            name="test_create_retrieve_delete",
            external_id=f"test_create_retrieve_delete_{RUN_UNIQUE_ID}",
            method="read_dial_gauge",
            input_schema=INPUT_SCHEMA_DATA_PROCESSING,
            description="Read dial gauge from an image using Cognite Vision gauge reader",
        )
        try:
            with contextlib.suppress(CogniteDuplicatedError):
                created = toolkit_client.robotics.data_postprocessing.create(data_processing)
                assert isinstance(created, DataPostProcessing)
                assert created.as_write().dump() == data_processing.dump()

            retrieved = toolkit_client.robotics.data_postprocessing.retrieve(data_processing.external_id)

            assert isinstance(retrieved, DataPostProcessing)
            assert retrieved.as_write().dump() == data_processing.dump()
        finally:
            toolkit_client.robotics.data_postprocessing.delete(data_processing.external_id)

        with pytest.raises(CogniteAPIError):
            toolkit_client.robotics.data_postprocessing.retrieve(data_processing.external_id)

    @pytest.mark.usefixtures("existing_data_processing")
    def test_list_data_postprocessing(self, toolkit_client: ToolkitClient) -> None:
        data_postprocessing = toolkit_client.robotics.data_postprocessing.list()
        assert isinstance(data_postprocessing, DataPostProcessingList)
        assert len(data_postprocessing) > 0

    @pytest.mark.usefixtures("existing_data_processing")
    def test_iterate_data_postprocessing(self, toolkit_client: ToolkitClient) -> None:
        for data_postprocessing in toolkit_client.robotics.data_postprocessing:
            assert isinstance(data_postprocessing, DataPostProcessing)
            break
        else:
            pytest.fail("No data processing found")

    def test_update_capability(
        self, toolkit_client: ToolkitClient, existing_data_processing: DataPostProcessing
    ) -> None:
        update = existing_data_processing.as_write()
        update.description = next(desc for desc in DESCRIPTIONS if desc != existing_data_processing.description)
        updated = toolkit_client.robotics.data_postprocessing.update(update)
        assert updated.description == update.description


class TestMapAPI:
    def test_create_retrieve_delete(self, toolkit_client: ToolkitClient) -> None:
        map_ = MapWrite(
            name="test_create_retrieve_delete",
            external_id=f"test_create_retrieve_delete_{RUN_UNIQUE_ID}",
            map_type="TWODMAP",
            description="2D map of the robotics lab",
            scale=1.0,
        )
        try:
            with contextlib.suppress(CogniteDuplicatedError):
                created = toolkit_client.robotics.maps.create(map_)
                assert isinstance(created, Map)
                assert created.as_write().dump() == map_.dump()

            retrieved = toolkit_client.robotics.maps.retrieve(map_.external_id)

            assert isinstance(retrieved, Map)
            assert retrieved.as_write().dump() == map_.dump()
        finally:
            toolkit_client.robotics.maps.delete(map_.external_id)

        with pytest.raises(CogniteAPIError):
            toolkit_client.robotics.maps.retrieve(map_.external_id)

    @pytest.mark.usefixtures("existing_map")
    def test_list_maps(self, toolkit_client: ToolkitClient) -> None:
        maps = toolkit_client.robotics.maps.list()
        assert isinstance(maps, MapList)
        assert len(maps) > 0

    @pytest.mark.usefixtures("existing_map")
    def test_iterate_maps(self, toolkit_client: ToolkitClient) -> None:
        for map_ in toolkit_client.robotics.maps:
            assert isinstance(map_, Map)
            break
        else:
            pytest.fail("No maps found")

    def test_update_map(self, toolkit_client: ToolkitClient, existing_map: Map) -> None:
        update = existing_map.as_write()
        update.description = next(desc for desc in DESCRIPTIONS if desc != existing_map.description)
        updated = toolkit_client.robotics.maps.update(update)
        assert updated.description == update.description


class TestLocationAPI:
    def test_create_retrieve_delete(self, toolkit_client: ToolkitClient) -> None:
        location = LocationWrite(
            name="test_create_retrieve_delete",
            external_id=f"test_create_retrieve_delete_{RUN_UNIQUE_ID}",
        )
        try:
            with contextlib.suppress(CogniteDuplicatedError):
                created = toolkit_client.robotics.locations.create(location)
                assert isinstance(created, Location)
                assert created.as_write().dump() == location.dump()

            retrieved = toolkit_client.robotics.locations.retrieve(location.external_id)

            assert isinstance(retrieved, Location)
            assert retrieved.as_write().dump() == location.dump()
        finally:
            toolkit_client.robotics.locations.delete(location.external_id)

        with pytest.raises(CogniteAPIError):
            toolkit_client.robotics.locations.retrieve(location.external_id)

    @pytest.mark.usefixtures("existing_location")
    def test_list_locations(self, toolkit_client: ToolkitClient) -> None:
        locations = toolkit_client.robotics.locations.list()
        assert isinstance(locations, LocationList)
        assert len(locations) > 0

    @pytest.mark.usefixtures("existing_location")
    def test_iterate_locations(self, toolkit_client: ToolkitClient) -> None:
        for location in toolkit_client.robotics.locations:
            assert isinstance(location, Location)
            break
        else:
            pytest.fail("No locations found")

    def test_update_location(self, toolkit_client: ToolkitClient, existing_location: Location) -> None:
        update = existing_location.as_write()
        update.description = next(desc for desc in DESCRIPTIONS if desc != existing_location.description)
        updated = toolkit_client.robotics.locations.update(update)
        assert updated.description == update.description


class TestFrameAPI:
    def test_create_retrieve_delete(self, toolkit_client: ToolkitClient, root_frame: Frame) -> None:
        frame = FrameWrite(
            name="test_create_retrieve_delete",
            external_id=f"test_create_retrieve_delete_{RUN_UNIQUE_ID}",
            transform=Transform(
                parent_frame_external_id=root_frame.external_id,
                translation=Point3D(0, 0, 0),
                orientation=Quaternion(0, 0, 0, 1),
            ),
        )
        try:
            with contextlib.suppress(CogniteDuplicatedError):
                created = toolkit_client.robotics.frames.create(frame)
                assert isinstance(created, Frame)
                assert created.as_write().dump() == frame.dump()

            retrieved = toolkit_client.robotics.frames.retrieve(frame.external_id)

            assert isinstance(retrieved, Frame)
            assert retrieved.as_write().dump() == frame.dump()
        finally:
            toolkit_client.robotics.frames.delete(frame.external_id)

        with pytest.raises(CogniteAPIError):
            toolkit_client.robotics.frames.retrieve(frame.external_id)

    @pytest.mark.usefixtures("existing_frame")
    def test_list_frames(self, toolkit_client: ToolkitClient) -> None:
        frames = toolkit_client.robotics.frames.list()
        assert isinstance(frames, FrameList)
        assert len(frames) > 0

    @pytest.mark.usefixtures("existing_frame")
    def test_iterate_frames(self, toolkit_client: ToolkitClient) -> None:
        for frame in toolkit_client.robotics.frames:
            assert isinstance(frame, Frame)
            break
        else:
            pytest.fail("No frames found")

    def test_update_frame(self, toolkit_client: ToolkitClient, existing_frame: Frame) -> None:
        update = existing_frame.as_write()
        update.name = next(name for name in FRAME_NAMES if name != existing_frame.name)
        updated = toolkit_client.robotics.frames.update(update)
        assert updated.name == update.name


DATA_HANDLING_SCHEMA_CAPABILITY = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "id": "robotics/schemas/0.1.0/data_handling/ptz",
    "type": "object",
    "properties": {
        "uploadInstructions": {
            "type": "object",
            "properties": {
                "image": {
                    "type": "object",
                    "properties": {
                        "method": {"const": "uploadFile"},
                        "parameters": {
                            "type": "object",
                            "properties": {"filenamePrefix": {"type": "string"}},
                            "required": ["filenamePrefix"],
                        },
                    },
                    "required": ["method", "parameters"],
                    "additionalProperties": False,
                }
            },
            "additionalProperties": False,
        }
    },
    "required": ["uploadInstructions"],
}

INPUT_SCHEMA_CAPABILITY = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "id": "robotics/schemas/0.1.0/capabilities/ptz",
    "title": "PTZ camera capability input",
    "type": "object",
    "properties": {
        "method": {"type": "string"},
        "parameters": {
            "type": "object",
            "properties": {
                "tilt": {"type": "number", "minimum": -90, "maximum": 90},
                "pan": {"type": "number", "minimum": -180, "maximum": 180},
                "zoom": {"type": "number", "minimum": 0, "maximum": 100},
            },
            "required": ["tilt", "pan", "zoom"],
        },
    },
    "required": ["method", "parameters"],
    "additionalProperties": False,
}

INPUT_SCHEMA_DATA_PROCESSING = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "id": "robotics/schemas/0.1.0/data_postprocessing/read_dial_gauge",
    "title": "Read dial gauge postprocessing input",
    "type": "object",
    "properties": {
        "image": {
            "type": "object",
            "properties": {
                "method": {"type": "string"},
                "parameters": {
                    "type": "object",
                    "properties": {
                        "unit": {"type": "string"},
                        "deadAngle": {"type": "number"},
                        "minLevel": {"type": "number"},
                        "maxLevel": {"type": "number"},
                    },
                },
            },
            "required": ["method", "parameters"],
            "additionalProperties": False,
        }
    },
    "additionalProperties": False,
}



from .agent_loaders import AgentLoader
from .auth_loaders import GroupAllScopedLoader, GroupLoader, SecurityCategoryLoader
from .classic_loaders import AssetLoader, EventLoader, SequenceLoader, SequenceRowLoader
from .data_organization_loaders import DataSetsLoader, LabelLoader
from .datamodel_loaders import (
    ContainerLoader,
    DataModelLoader,
    EdgeLoader,
    GraphQLLoader,
    NodeLoader,
    SpaceLoader,
    ViewLoader,
)
from .extraction_pipeline_loaders import ExtractionPipelineConfigLoader, ExtractionPipelineLoader
from .fieldops_loaders import InfieldV1Loader
from .file_loader import CogniteFileLoader, FileMetadataLoader
from .function_loaders import FunctionLoader, FunctionScheduleLoader
from .group_scoped_loader import GroupResourceScopedLoader
from .hosted_extractors import (
    HostedExtractorDestinationLoader,
    HostedExtractorJobLoader,
    HostedExtractorMappingLoader,
    HostedExtractorSourceLoader,
)
from .industrial_tool_loaders import StreamlitLoader
from .location_loaders import LocationFilterLoader
from .raw_loaders import RawDatabaseLoader, RawTableLoader
from .relationship_loader import RelationshipLoader
from .robotics_loaders import (
    RobotCapabilityLoader,
    RoboticFrameLoader,
    RoboticLocationLoader,
    RoboticMapLoader,
    RoboticsDataPostProcessingLoader,
)
from .three_d_model_loaders import ThreeDModelLoader
from .timeseries_loaders import DatapointSubscriptionLoader, TimeSeriesLoader
from .transformation_loaders import TransformationLoader, TransformationNotificationLoader, TransformationScheduleLoader
from .workflow_loaders import WorkflowLoader, WorkflowTriggerLoader, WorkflowVersionLoader

__all__ = [
    "AgentLoader",
    "AssetLoader",
    "CogniteFileLoader",
    "ContainerLoader",
    "DataModelLoader",
    "DataSetsLoader",
    "DatapointSubscriptionLoader",
    "EdgeLoader",
    "EventLoader",
    "ExtractionPipelineConfigLoader",
    "ExtractionPipelineLoader",
    "FileMetadataLoader",
    "FunctionLoader",
    "FunctionScheduleLoader",
    "GraphQLLoader",
    "GroupAllScopedLoader",
    "GroupLoader",
    "GroupResourceScopedLoader",
    "HostedExtractorDestinationLoader",
    "HostedExtractorJobLoader",
    "HostedExtractorMappingLoader",
    "HostedExtractorSourceLoader",
    "InfieldV1Loader",
    "LabelLoader",
    "LocationFilterLoader",
    "NodeLoader",
    "RawDatabaseLoader",
    "RawTableLoader",
    "RelationshipLoader",
    "RobotCapabilityLoader",
    "RoboticFrameLoader",
    "RoboticLocationLoader",
    "RoboticMapLoader",
    "RoboticsDataPostProcessingLoader",
    "SecurityCategoryLoader",
    "SequenceLoader",
    "SequenceRowLoader",
    "SpaceLoader",
    "StreamlitLoader",
    "ThreeDModelLoader",
    "TimeSeriesLoader",
    "TransformationLoader",
    "TransformationNotificationLoader",
    "TransformationScheduleLoader",
    "ViewLoader",
    "WorkflowLoader",
    "WorkflowTriggerLoader",
    "WorkflowVersionLoader",
]



from __future__ import annotations

import collections.abc
import enum
import sys
from collections.abc import Iterable, Mapping, Sequence
from typing import Any, List, Literal, Union  # noqa: UP035

import pytest
from _pytest.mark import ParameterSet
from cognite.client.data_classes.data_modeling import DirectRelationReference, NodeId
from cognite.client.data_classes.data_modeling.instances import PropertyValue

from cognite_toolkit._cdf_tk._parameters.type_hint import TypeHint


class Action(enum.Enum):
    Read = "READ"
    Execute = "EXECUTE"
    List = "LIST"


def type_hint_test_cases() -> Iterable[ParameterSet]:
    # "raw, types, is_base_type, is_nullable, is_class, is_dict_type, is_list_type",
    yield pytest.param(str, ["str"], True, False, False, False, False, id="str")
    yield pytest.param(Literal["a", "b"], ["str"], True, False, False, False, False, id="Literal")
    yield pytest.param(dict[str, int], ["dict"], False, False, False, True, False, id="dict")
    yield pytest.param(Union[str, int], ["str", "int"], True, False, False, False, False, id="Union")
    yield pytest.param(dict, ["dict"], False, False, False, True, False, id="dict without type hints")
    yield pytest.param(Any, ["unknown"], False, False, False, False, False, id="Any")
    yield pytest.param(Action, ["str"], True, False, True, False, False, id="Enum")
    yield pytest.param(
        Union[
            str,
            int,
            float,
            bool,
            dict,
            List[str],  # noqa: UP006
            List[int],  # noqa: UP006
            List[float],  # noqa: UP006
            List[bool],  # noqa: UP006
            List[dict],  # noqa: UP006
            NodeId,
            DirectRelationReference,
        ],
        ["str", "int", "float", "bool", "dict", "list"],
        True,
        False,
        True,
        True,
        True,
        id="Union with almost all possible types",
    )
    if sys.version_info >= (3, 10):
        yield pytest.param(Sequence[int], ["list"], False, False, False, False, True, id="Sequence")
        yield pytest.param(Mapping[str, PropertyValue], ["dict"], False, False, False, True, False, id="Mapping")
        yield pytest.param(
            collections.abc.Mapping[str, PropertyValue], ["dict"], False, False, False, True, False, id="ABC Mapping"
        )
        yield pytest.param(str | None, ["str"], True, True, False, False, False, id="str | None")
        yield pytest.param(
            str | int | bool, ["str", "int", "bool"], True, False, False, False, False, id="str | int | bool"
        )
        yield pytest.param(list[str] | None, ["list"], False, True, False, False, True, id="list | None")
        yield pytest.param(
            list[str] | dict[str, int] | None, ["list", "dict"], False, True, False, True, True, id="list | dict | None"
        )


@pytest.mark.skipif(sys.version_info[:2] == (3, 10), reason="Fails for Python 3.10, tech debt to fix")
class TestTypeHint:
    @pytest.mark.parametrize(
        "raw, types, is_base_type, is_nullable, is_class, is_dict_type, is_list_type",
        list(type_hint_test_cases()),
    )
    def test_type_hint(
        self,
        raw: Any,
        types: list[str],
        is_base_type: bool,
        is_nullable: bool,
        is_class: bool,
        is_dict_type: bool,
        is_list_type: bool,
    ):
        hint = TypeHint(raw)
        assert set(hint.types) == set(types)
        assert hint.is_base_type == is_base_type
        assert hint.is_nullable == is_nullable
        assert hint.is_user_defined_class == is_class
        assert hint.is_dict_type == is_dict_type
        assert hint.is_list_type == is_list_type



from __future__ import annotations

from dataclasses import dataclass
from functools import cached_property
from pathlib import Path
from typing import Any, ClassVar

import yaml

from cognite_toolkit import _version
from cognite_toolkit._cdf_tk.cdf_toml import CDFToml
from cognite_toolkit._cdf_tk.constants import DEFAULT_ENV
from cognite_toolkit._cdf_tk.loaders import ResourceTypes
from cognite_toolkit._cdf_tk.loaders._base_loaders import T_ID
from cognite_toolkit._cdf_tk.utils import (
    safe_write,
    tmp_build_directory,
)
from cognite_toolkit._cdf_tk.utils.file import yaml_safe_dump

from ._base import ConfigCore
from ._build_variables import BuildVariables
from ._built_modules import BuiltModule, BuiltModuleList
from ._built_resources import BuiltFullResourceList
from ._config_yaml import BuildConfigYAML
from ._module_directories import ModuleDirectories


@dataclass
class ModulesInfo:
    version: str
    modules: BuiltModuleList

    @classmethod
    def load(cls, data: dict[str, Any]) -> ModulesInfo:
        return cls(
            version=data["version"],
            modules=BuiltModuleList([BuiltModule.load(module_data) for module_data in data["modules"]]),
        )

    def dump(self) -> dict[str, Any]:
        return {
            "version": self.version,
            "modules": [module.dump() for module in self.modules],
        }


@dataclass
class BuildInfo(ConfigCore):
    filename: ClassVar[str] = "build_info.{build_env}.yaml"
    top_warning: ClassVar[str] = "# DO NOT MODIFY THIS FILE MANUALLY. IT IS AUTO-GENERATED BY THE COGNITE TOOLKIT."
    modules: ModulesInfo

    @classmethod
    def load(cls, data: dict[str, Any], build_env: str, filepath: Path) -> BuildInfo:
        return cls(filepath, ModulesInfo.load(data["modules"]))

    @classmethod
    def rebuild(
        cls, organization_dir: Path, build_env: str | None, needs_rebuild: set[Path] | None = None
    ) -> BuildInfo:
        # To avoid circular imports
        # Ideally, this class should be in a separate module
        from cognite_toolkit._cdf_tk.commands.build_cmd import BuildCommand

        with tmp_build_directory() as build_dir:
            cdf_toml = CDFToml.load()
            if build_env is None:
                config = BuildConfigYAML.load_default(organization_dir)
            else:
                config = BuildConfigYAML.load_from_directory(organization_dir, build_env)
            config.set_environment_variables()
            # Todo Remove once the new modules in `_cdf_tk/prototypes/_packages` are finished.
            config.variables.pop("_cdf_tk", None)
            if needs_rebuild is None:
                # Use path syntax to select all modules in the source directory
                config.environment.selected = [Path("")]
            else:
                # Use path syntax to select only the modules that need to be rebuilt
                config.environment.selected = list(needs_rebuild)
            build = BuildCommand(silent=True, skip_tracking=True).build_config(
                build_dir=build_dir,
                organization_dir=organization_dir,
                config=config,
                packages=cdf_toml.modules.packages,
                clean=True,
                verbose=False,
                progress_bar=True,
            )

        new_build = cls(
            filepath=organization_dir / cls.get_filename(build_env or DEFAULT_ENV),
            modules=ModulesInfo(version=_version.__version__, modules=build),
        )
        if needs_rebuild is not None and (existing := cls._get_existing(organization_dir, build_env or DEFAULT_ENV)):
            # Merge the existing modules with the new modules
            new_modules_by_path = {module.location.path: module for module in new_build.modules.modules}
            existing_modules_by_path = {module.location.path: module for module in existing.modules.modules}
            all_module_paths = set(new_modules_by_path) | set(existing_modules_by_path)

            module_list = BuiltModuleList(
                [
                    new_modules_by_path[path] if path in new_modules_by_path else existing_modules_by_path[path]
                    for path in all_module_paths
                ]
            )
            new_build.modules.modules = module_list

        new_build.dump_to_file()
        return new_build

    @classmethod
    def _get_existing(cls, organization_dir: Path, build_env: str) -> BuildInfo | None:
        try:
            existing = cls.load_from_directory(organization_dir, build_env)
        except FileNotFoundError:
            return None
        if existing.modules.version != _version.__version__:
            return None
        return existing

    def dump(self) -> dict[str, Any]:
        return {
            "modules": self.modules.dump(),
        }

    def dump_to_file(self) -> None:
        dumped = self.dump()
        # Avoid dumping pointer references: https://stackoverflow.com/questions/51272814/python-yaml-dumping-pointer-references
        yaml.Dumper.ignore_aliases = lambda *args: True  # type: ignore[method-assign]
        content = yaml_safe_dump(dumped)
        content = f"{self.top_warning}\n{content}"
        safe_write(self.filepath, content)

    def compare_modules(
        self,
        current_modules: ModuleDirectories,
        current_variables: BuildVariables,
        resource_dirs: set[str] | None = None,
    ) -> set[Path]:
        current_module_by_path = {module.relative_path: module for module in current_modules}
        cached_module_by_path_and_iteration = {
            (module.location.path, module.iteration): module for module in self.modules.modules
        }
        needs_rebuild = set()
        for path, current_module in current_module_by_path.items():
            if resource_dirs is not None and all(
                resource_dir not in current_module.resource_directories for resource_dir in resource_dirs
            ):
                # The module does not contain any of the specified resources, so it does not need to be rebuilt.
                continue

            if (path, None) not in cached_module_by_path_and_iteration and (
                path,
                1,
            ) not in cached_module_by_path_and_iteration:
                needs_rebuild.add(path)
                continue
            for no, current_module_variables in enumerate(current_variables.get_module_variables(current_module), 1):
                cached_module = cached_module_by_path_and_iteration[(path, no)]
                if current_module.hash != cached_module.location.hash:
                    needs_rebuild.add(path)
                    continue
                if set(current_module_variables) != set(cached_module.build_variables):
                    needs_rebuild.add(path)
        return needs_rebuild


class ModuleResources:
    """This class is used to retrieve resource information from the build info.

    It is responsible for ensuring that the build info is up-to-date with the
    latest changes in the source directory.
    """

    def __init__(self, organization_dir: Path, build_env: str | None) -> None:
        self._organization_dir = organization_dir
        self._build_env = build_env or DEFAULT_ENV
        self._build_info: BuildInfo
        try:
            self._build_info = BuildInfo.load_from_directory(organization_dir, build_env or DEFAULT_ENV)
            self._has_rebuilt = False
        except (FileNotFoundError, KeyError):
            # FileNotFound = Not run before.
            # KeyError = Version mismatch/Changed format
            self._build_info = BuildInfo.rebuild(organization_dir, build_env)
            self._has_rebuilt = True

    @cached_property
    def _current_modules(self) -> ModuleDirectories:
        return ModuleDirectories.load(self._organization_dir, {Path("")})

    @cached_property
    def _current_variables(self) -> BuildVariables:
        config_yaml = BuildConfigYAML.load_from_directory(self._organization_dir, self._build_env)
        return BuildVariables.load_raw(
            config_yaml.variables, self._current_modules.available_paths, self._current_modules.selected.available_paths
        )

    def list_resources(
        self, id_type: type[T_ID], resource_dir: ResourceTypes, kind: str
    ) -> BuiltFullResourceList[T_ID]:
        if not self._has_rebuilt:
            if needs_rebuild := self._build_info.compare_modules(
                self._current_modules, self._current_variables, {resource_dir}
            ):
                self._build_info = BuildInfo.rebuild(self._organization_dir, self._build_env, needs_rebuild)
        return self._build_info.modules.modules.get_resources(id_type, resource_dir, kind)

    def list(self) -> BuiltModuleList:
        # Check if the build info is up to date
        if not self._has_rebuilt:
            if needs_rebuild := self._build_info.compare_modules(self._current_modules, self._current_variables):
                self._build_info = BuildInfo.rebuild(self._organization_dir, self._build_env, needs_rebuild)
            self._has_rebuilt = True
        return self._build_info.modules.modules



# Copyright 2023 Cognite AS
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from collections.abc import Hashable, Iterable, Sequence
from datetime import date, datetime
from functools import lru_cache
from typing import Any, cast, final

from cognite.client.data_classes import (
    FileMetadata,
    FileMetadataList,
    FileMetadataWrite,
    FileMetadataWriteList,
)
from cognite.client.data_classes.capabilities import (
    Capability,
    DataModelInstancesAcl,
    FilesAcl,
)
from cognite.client.data_classes.data_modeling import NodeApplyResultList, NodeId, ViewId
from cognite.client.exceptions import CogniteAPIError
from cognite.client.utils._time import convert_data_modelling_timestamp
from cognite.client.utils.useful_types import SequenceNotStr
from rich import print

from cognite_toolkit._cdf_tk._parameters import ANY_INT, ANY_STR, ANYTHING, ParameterSpec, ParameterSpecSet
from cognite_toolkit._cdf_tk.client.data_classes.extendable_cognite_file import (
    ExtendableCogniteFile,
    ExtendableCogniteFileApply,
    ExtendableCogniteFileApplyList,
    ExtendableCogniteFileList,
)
from cognite_toolkit._cdf_tk.exceptions import (
    ToolkitRequiredValueError,
)
from cognite_toolkit._cdf_tk.loaders._base_loaders import ResourceContainerLoader, ResourceLoader
from cognite_toolkit._cdf_tk.resource_classes import FileMetadataYAML
from cognite_toolkit._cdf_tk.utils import (
    in_dict,
)
from cognite_toolkit._cdf_tk.utils.diff_list import diff_list_hashable, diff_list_identifiable, dm_identifier

from .auth_loaders import GroupAllScopedLoader, SecurityCategoryLoader
from .classic_loaders import AssetLoader
from .data_organization_loaders import DataSetsLoader, LabelLoader
from .datamodel_loaders import SpaceLoader, ViewLoader


@final
class FileMetadataLoader(
    ResourceContainerLoader[str, FileMetadataWrite, FileMetadata, FileMetadataWriteList, FileMetadataList]
):
    item_name = "file contents"
    folder_name = "files"
    filename_pattern = (
        # Matches all yaml files except file names whose stem ends with `.CogniteFile` or `File`.
        r"(?i)^(?!.*(?:File|CogniteFile)$).*$"
    )
    resource_cls = FileMetadata
    resource_write_cls = FileMetadataWrite
    list_cls = FileMetadataList
    list_write_cls = FileMetadataWriteList
    yaml_cls = FileMetadataYAML
    kind = "FileMetadata"
    dependencies = frozenset({DataSetsLoader, GroupAllScopedLoader, LabelLoader, AssetLoader})

    _doc_url = "Files/operation/initFileUpload"

    @property
    def display_name(self) -> str:
        return "file metadata"

    @classmethod
    def get_required_capability(
        cls, items: Sequence[FileMetadataWrite] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = [FilesAcl.Action.Read] if read_only else [FilesAcl.Action.Read, FilesAcl.Action.Write]

        scope: FilesAcl.Scope.All | FilesAcl.Scope.DataSet = FilesAcl.Scope.All()  # type: ignore[valid-type]
        if items:
            if data_set_ids := {item.data_set_id for item in items if item.data_set_id}:
                scope = FilesAcl.Scope.DataSet(list(data_set_ids))

        return FilesAcl(actions, scope)  # type: ignore[arg-type]

    @classmethod
    def get_id(cls, item: FileMetadata | FileMetadataWrite | dict) -> str:
        if isinstance(item, dict):
            return item["externalId"]
        if item.external_id is None:
            raise ToolkitRequiredValueError("FileMetadata must have external_id set.")
        return item.external_id

    @classmethod
    def get_internal_id(cls, item: FileMetadata | dict) -> int:
        if isinstance(item, dict):
            return item["id"]
        return item.id

    @classmethod
    def dump_id(cls, id: str) -> dict[str, Any]:
        return {"externalId": id}

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceLoader], Hashable]]:
        if "dataSetExternalId" in item:
            yield DataSetsLoader, item["dataSetExternalId"]
        if "securityCategoryNames" in item:
            for security_category in item["securityCategoryNames"]:
                yield SecurityCategoryLoader, security_category
        if "labels" in item:
            for label in item["labels"]:
                if isinstance(label, dict):
                    yield LabelLoader, label["externalId"]
                elif isinstance(label, str):
                    yield LabelLoader, label
        for asset_external_id in item.get("assetExternalIds", []):
            yield AssetLoader, asset_external_id

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> FileMetadataWrite:
        if resource.get("dataSetExternalId") is not None:
            ds_external_id = resource.pop("dataSetExternalId")
            resource["dataSetId"] = self.client.lookup.data_sets.id(ds_external_id, is_dry_run)
        if security_categories_names := resource.pop("securityCategoryNames", []):
            security_categories = self.client.lookup.security_categories.id(security_categories_names, is_dry_run)
            resource["securityCategories"] = security_categories
        if "assetExternalIds" in resource:
            resource["assetIds"] = self.client.lookup.assets.id(resource["assetExternalIds"], is_dry_run)
        return FileMetadataWrite._load(resource)

    def dump_resource(self, resource: FileMetadata, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_write().dump()
        if ds_id := dumped.pop("dataSetId", None):
            dumped["dataSetExternalId"] = self.client.lookup.data_sets.external_id(ds_id)
        if security_categories := dumped.pop("securityCategories", []):
            dumped["securityCategoryNames"] = self.client.lookup.security_categories.external_id(security_categories)
        if asset_ids := dumped.pop("assetIds", []):
            dumped["assetExternalIds"] = self.client.lookup.assets.external_id(asset_ids)
        return dumped

    def create(self, items: FileMetadataWriteList) -> FileMetadataList:
        created = FileMetadataList([])
        for meta in items:
            try:
                created.append(self.client.files.create(meta))
            except CogniteAPIError as e:
                if e.code == 409:
                    print(f"  [bold yellow]WARNING:[/] File {meta.external_id} already exists, skipping upload.")
        return created

    def retrieve(self, ids: SequenceNotStr[str]) -> FileMetadataList:
        return self.client.files.retrieve_multiple(external_ids=ids, ignore_unknown_ids=True)

    def update(self, items: FileMetadataWriteList) -> FileMetadataList:
        return self.client.files.update(items, mode="replace")

    def delete(self, ids: str | int | SequenceNotStr[str | int] | None) -> int:
        internal_ids, external_ids = self._split_ids(ids)
        self.client.files.delete(id=internal_ids, external_id=external_ids)
        return len(cast(SequenceNotStr[str], ids))

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[FileMetadata]:
        return iter(self.client.files(data_set_external_ids=[data_set_external_id] if data_set_external_id else None))

    def count(self, ids: SequenceNotStr[str]) -> int:
        return sum(
            1
            for meta in self.client.files.retrieve_multiple(external_ids=list(ids), ignore_unknown_ids=True)
            if meta.uploaded
        )

    def drop_data(self, ids: SequenceNotStr[str]) -> int:
        existing = self.client.files.retrieve_multiple(external_ids=list(ids), ignore_unknown_ids=True)
        # File and FileMetadata is tightly coupled, so we need to delete the metadata and recreate it
        # without the source set to delete the file.
        deleted_files = self.delete(existing.as_external_ids())
        self.create(existing.as_write())
        return deleted_files

    @classmethod
    @lru_cache(maxsize=1)
    def get_write_cls_parameter_spec(cls) -> ParameterSpecSet:
        spec = super().get_write_cls_parameter_spec()
        # Added by toolkit
        spec.add(ParameterSpec(("dataSetExternalId",), frozenset({"str"}), is_required=False, _is_nullable=False))
        spec.discard(ParameterSpec(("dataSetId",), frozenset({"int"}), is_required=False, _is_nullable=False))
        spec.add(ParameterSpec(("securityCategoryNames",), frozenset({"list"}), is_required=False, _is_nullable=False))

        spec.add(
            ParameterSpec(("securityCategoryNames", ANY_STR), frozenset({"str"}), is_required=False, _is_nullable=False)
        )
        spec.add(ParameterSpec(("assetExternalIds",), frozenset({"list"}), is_required=False, _is_nullable=False))
        spec.add(
            ParameterSpec(("assetExternalIds", ANY_INT), frozenset({"int"}), is_required=False, _is_nullable=False)
        )
        spec.discard(ParameterSpec(("assetIds",), frozenset({"list"}), is_required=False, _is_nullable=False))
        spec.discard(ParameterSpec(("assetIds", ANY_INT), frozenset({"str"}), is_required=False, _is_nullable=False))

        return spec


@final
class CogniteFileLoader(
    ResourceContainerLoader[
        NodeId,
        ExtendableCogniteFileApply,
        ExtendableCogniteFile,
        ExtendableCogniteFileApplyList,
        ExtendableCogniteFileList,
    ]
):
    template_pattern = "$FILENAME"
    item_name = "file contents"
    folder_name = "files"
    filename_pattern = r"^.*CogniteFile"  # Matches all yaml files whose stem ends with 'CogniteFile'.
    kind = "CogniteFile"
    resource_cls = ExtendableCogniteFile
    resource_write_cls = ExtendableCogniteFileApply
    list_cls = ExtendableCogniteFileList
    list_write_cls = ExtendableCogniteFileApplyList
    dependencies = frozenset({GroupAllScopedLoader, SpaceLoader, ViewLoader})

    _doc_url = "Files/operation/initFileUpload"

    @property
    def display_name(self) -> str:
        return "cognite files"

    @classmethod
    def get_id(cls, item: ExtendableCogniteFile | ExtendableCogniteFileApply | dict) -> NodeId:
        if isinstance(item, dict):
            if missing := tuple(k for k in {"space", "externalId"} if k not in item):
                # We need to raise a KeyError with all missing keys to get the correct error message.
                raise KeyError(*missing)
            return NodeId(space=item["space"], external_id=item["externalId"])
        return item.as_id()

    @classmethod
    def dump_id(cls, id: NodeId) -> dict[str, Any]:
        return id.dump(include_instance_type=False)

    @classmethod
    def get_required_capability(
        cls, items: Sequence[ExtendableCogniteFileApply] | None, read_only: bool
    ) -> list[Capability]:
        if not items and items is not None:
            return []

        file_actions = [FilesAcl.Action.Read] if read_only else [FilesAcl.Action.Read, FilesAcl.Action.Write]
        instance_actions = (
            [DataModelInstancesAcl.Action.Read]
            if read_only
            else [DataModelInstancesAcl.Action.Read, DataModelInstancesAcl.Action.Write]
        )

        scope: DataModelInstancesAcl.Scope.All | DataModelInstancesAcl.Scope.SpaceID = DataModelInstancesAcl.Scope.All()  # type: ignore[valid-type]
        if items:
            if spaces := {item.space for item in items}:
                scope = DataModelInstancesAcl.Scope.SpaceID(list(spaces))
        return [
            FilesAcl(file_actions, FilesAcl.Scope.All()),
            DataModelInstancesAcl(
                instance_actions,  # type: ignore[valid-type]
                scope,  # type: ignore[arg-type]
            ),
        ]

    def dump_resource(self, resource: ExtendableCogniteFile, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_write().dump(context="local")
        local = local or {}
        if "existingVersion" not in local:
            # Existing version is typically not set when creating nodes, but we get it back
            # when we retrieve the node from the server.
            dumped.pop("existingVersion", None)
        for key in list(dumped.keys()):
            value = dumped[key]
            if key not in local:
                if value is None:
                    dumped.pop(key)
                continue
            local_value = local[key]
            if isinstance(local_value, datetime) and isinstance(value, str):
                dumped[key] = convert_data_modelling_timestamp(value)
            elif isinstance(local_value, date) and isinstance(value, str):
                dumped[key] = date.fromisoformat(value)

        if "nodeSource" in local:
            dumped["nodeSource"] = local["nodeSource"]
        if dumped.get("instanceType") == "node" and "instanceType" not in local:
            dumped.pop("instanceType")
        return dumped

    def diff_list(
        self, local: list[Any], cdf: list[Any], json_path: tuple[str | int, ...]
    ) -> tuple[dict[int, int], list[int]]:
        if json_path == ("tags",):
            return diff_list_hashable(local, cdf)
        elif json_path[0] in ("assets", "category"):
            return diff_list_identifiable(local, cdf, get_identifier=dm_identifier)
        return super().diff_list(local, cdf, json_path)

    def create(self, items: ExtendableCogniteFileApplyList) -> NodeApplyResultList:
        created = self.client.data_modeling.instances.apply(
            nodes=items, replace=False, skip_on_version_conflict=True, auto_create_direct_relations=True
        )
        return created.nodes

    def retrieve(self, ids: SequenceNotStr[NodeId]) -> ExtendableCogniteFileList:
        # Todo: Problem, if you extend the CogniteFiles with a custom view, we need to know
        #   the ID of the custom view to retrieve data from it. This is not possible with the current
        #   structure. Need to reconsider how to handle this.
        items = self.client.data_modeling.instances.retrieve_nodes(  # type: ignore[call-overload]
            nodes=ids,
            node_cls=ExtendableCogniteFile,
        )
        return ExtendableCogniteFileList(items)

    def update(self, items: ExtendableCogniteFileApplyList) -> NodeApplyResultList:
        updated = self.client.data_modeling.instances.apply(nodes=items, replace=True)
        return updated.nodes

    def delete(self, ids: SequenceNotStr[NodeId]) -> int:
        try:
            deleted = self.client.data_modeling.instances.delete(nodes=list(ids))
        except CogniteAPIError as e:
            if "not exist" in e.message and "space" in e.message.lower():
                return 0
            raise e
        return len(deleted.nodes)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[ExtendableCogniteFile]:
        # We do not have a way to know the source of the file, so we cannot filter on that.
        return []

    def count(self, ids: SequenceNotStr[NodeId]) -> int:
        return sum(
            [
                bool(n.is_uploaded or False)
                for n in self.client.data_modeling.instances.retrieve_nodes(nodes=ids, node_cls=ExtendableCogniteFile)  # type: ignore[call-overload]
            ]
        )

    def drop_data(self, ids: SequenceNotStr[NodeId]) -> int:
        existing_meta = self.client.files.retrieve_multiple(instance_ids=list(ids), ignore_unknown_ids=True)
        existing_node = self.retrieve(ids)

        # File and FileMetadata is tightly coupled, so we need to delete the metadata and recreate it
        # without the source set to delete the file.
        self.client.files.delete(id=existing_meta.as_ids())
        self.create(existing_node.as_write())
        return len(existing_meta)

    @classmethod
    @lru_cache(maxsize=1)
    def get_write_cls_parameter_spec(cls) -> ParameterSpecSet:
        spec = super().get_write_cls_parameter_spec()
        # Removed by the SDK
        spec.add(ParameterSpec(("instanceType",), frozenset({"str"}), is_required=False, _is_nullable=False))
        # Sources are used when writing to the API.
        spec.add(ParameterSpec(("sources",), frozenset({"list"}), is_required=False, _is_nullable=False))
        spec.add(ParameterSpec(("sources", ANYTHING), frozenset({"list"}), is_required=False, _is_nullable=False))
        return spec

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceLoader], Hashable]]:
        """Returns all items that this item requires.

        For example, a TimeSeries requires a DataSet, so this method would return the
        DatasetLoader and identifier of that dataset.
        """
        if "space" in item:
            yield SpaceLoader, item["space"]
        if "nodeSource" in item:
            if in_dict(("space", "externalId", "type"), item["nodeSource"]):
                yield ViewLoader, ViewId.load(item["nodeSource"])



from collections.abc import Sequence
from pathlib import Path
from typing import Any

import pytest
import typer
from _pytest.mark import ParameterSet
from _pytest.monkeypatch import MonkeyPatch
from questionary import Choice

from cognite_toolkit._cdf_tk.commands import BuildCommand, ModulesCommand
from cognite_toolkit._cdf_tk.constants import BUILTIN_MODULES_PATH
from cognite_toolkit._cdf_tk.data_classes import Package, Packages
from cognite_toolkit._cdf_tk.tk_warnings import DuplicatedItemWarning, TemplateVariableWarning
from tests.test_unit.utils import MockQuestionary


def get_packages() -> list[ParameterSet]:
    packages = Packages.load(BUILTIN_MODULES_PATH)
    # - The Bootcamp package has intentionally warnings that is part of the learning experience.
    # - Examples and sourcesystems are tested separately, in that each example is tested individually as they
    # - Custom is just scaffolding and should never issue warnings.
    # should be independent of each other.
    packages = (
        package
        for package in packages.values()
        if package.name not in ["bootcamp", "sourcesystem", "industrial_tools", "contextualization", "custom"]
    )
    return [pytest.param(package, id=package.name) for package in sorted(packages, key=lambda p: p.name)]


@pytest.mark.parametrize("package", get_packages())
def test_build_packages_without_warnings(
    package: Package, tmp_path: Path, build_tmp_path: Path, monkeypatch: MonkeyPatch
) -> None:
    organization_dir = tmp_path

    module_cmd = ModulesCommand(silent=True, skip_tracking=True)

    def select_package(choices: Sequence[Choice]) -> Any:
        return next(choice.value for choice in choices if choice.value.name == package.name)

    answers = [
        select_package,
        False,
        True,
        ["dev"],
    ]
    if package.can_cherry_pick:
        answers.insert(1, MockQuestionary.select_all)

    if package.name == "quickstart":
        # Skip downloading example data.
        answers.append(False)

    with MockQuestionary(ModulesCommand.__module__, monkeypatch, answers), pytest.raises(typer.Exit) as exc_info:
        module_cmd.init(organization_dir, clean=True)

    assert exc_info.value.exit_code == 0

    build_cmd = BuildCommand(silent=True, skip_tracking=True)

    monkeypatch.setenv("CDF_PROJECT", "<my-project-dev>")
    build_cmd.execute(
        verbose=False,
        build_dir=build_tmp_path,
        organization_dir=organization_dir,
        build_env_name="dev",
        no_clean=False,
        client=None,
        selected=None,
    )

    # TemplateVariableWarning is when <change_me> is not replaced in the config file.
    # This is expected to be replaced by the users, and will thus raise when we run a fully automated test.
    warnings = [warning for warning in build_cmd.warning_list if not isinstance(warning, TemplateVariableWarning)]
    if package.name == "quickstart":
        # QuickStart combines modules that have the same spaces/datasets etc.
        warnings = [warning for warning in warnings if not isinstance(warning, DuplicatedItemWarning)]

    assert not warnings, f"{len(warnings)} warnings found: {warnings}"


def get_individual_modules() -> list[str]:
    packages = Packages.load(BUILTIN_MODULES_PATH)

    for package_name in ["sourcesystem"]:
        modules = packages[package_name]
        for module_name in sorted(modules.module_names):
            yield package_name, module_name


@pytest.mark.parametrize("package, module_name", get_individual_modules())
def test_build_individual_module(
    package: str, module_name: str, tmp_path: Path, build_tmp_path: Path, monkeypatch
) -> None:
    organization_dir = tmp_path

    module_cmd = ModulesCommand(silent=True, skip_tracking=True)

    def select_package(choices: Sequence[Choice]) -> Any:
        return next(choice.value for choice in choices if choice.value.name == package)

    def select_module(choices: Sequence[Choice]) -> Any:
        return [next(choice.value for choice in choices if choice.value.name == module_name)]

    answers = [
        select_package,
        select_module,
        False,
        True,
        ["dev"],
    ]
    if package == "sourcesystem":
        # Answer no to download example data.
        answers.append(False)

    with MockQuestionary(ModulesCommand.__module__, monkeypatch, answers), pytest.raises(typer.Exit) as exc_info:
        module_cmd.init(organization_dir, clean=True)

    assert exc_info.value.exit_code == 0

    build_cmd = BuildCommand(silent=True, skip_tracking=True)

    monkeypatch.setenv("CDF_PROJECT", "<my-project-dev>")
    build_cmd.execute(
        verbose=False,
        build_dir=build_tmp_path,
        organization_dir=organization_dir,
        build_env_name="dev",
        no_clean=False,
        client=None,
        selected=None,
    )

    # TemplateVariableWarning is when <change_me> is not replaced in the config file.
    # This is expected to be replaced by the users, and will thus raise when we run a fully automated test.
    warnings = [warning for warning in build_cmd.warning_list if not isinstance(warning, TemplateVariableWarning)]

    assert not warnings, f"{len(warnings)} warnings found: {warnings}"


def test_all_modules_cdf_prefixed() -> None:
    packages = Packages.load(BUILTIN_MODULES_PATH)
    missing_cdf_prefix = {
        module.name
        for package in packages.values()
        # Bootcamp has special structure
        if package.name not in {"bootcamp", "custom"}
        for module in package.modules
        if not module.name.startswith("cdf_")
    }

    assert not missing_cdf_prefix, f"Modules missing cdf_ prefix: {missing_cdf_prefix}"


def test_no_builtin_duplicates(organization_dir: Path, build_tmp_path: Path) -> None:
    cmd = BuildCommand(silent=True)

    modules = Path("modules")
    cmd.execute(
        verbose=False,
        organization_dir=organization_dir,
        build_dir=build_tmp_path,
        build_env_name="dev",
        no_clean=False,
        client=None,
        selected=[
            modules / "cdf_ingestion",
            modules / "common",
            modules / "contextualization",
            modules / "industrial_tools",
            modules / "models",
            modules / "sourcesystem",
        ],
    )

    duplicate_warning = [warning for warning in cmd.warning_list if isinstance(warning, DuplicatedItemWarning)]

    assert not duplicate_warning, f"{len(duplicate_warning)} duplicate warnings found: {duplicate_warning}"


def test_all_extra_resources_exists() -> None:
    packages = Packages.load(BUILTIN_MODULES_PATH)
    missing_resources = {
        extra: module.name
        for package in packages.values()
        for module in package.modules
        if module.definition
        for extra in module.definition.extra_resources
        if not (BUILTIN_MODULES_PATH / extra).exists()
    }
    missing_by_module = {v: k.as_posix() for k, v in missing_resources.items()}

    assert not missing_resources, f"Modules missing resources: {missing_by_module}"



import platform
import random
import sys

# This is needed as we run tests for two different versions of Python in parallel.
# The platform.system() is not used, but is here in case we start testing on Windows as well.
# The random number is to avoid conflicts when running tests in parallel (for example, two PRs).
RUN_UNIQUE_ID = f"{platform.system()}_{sys.version_info.major}_{sys.version_info.minor}_{random.randint(0, 10)!s}"



from collections.abc import Hashable
from pathlib import Path
from unittest.mock import MagicMock

import pytest
import yaml
from _pytest.monkeypatch import MonkeyPatch
from cognite.client.data_classes import Transformation
from cognite.client.data_classes import data_modeling as dm

from cognite_toolkit._cdf_tk.client.data_classes.raw import RawDatabase, RawTable
from cognite_toolkit._cdf_tk.client.testing import monkeypatch_toolkit_client
from cognite_toolkit._cdf_tk.loaders import (
    DataModelLoader,
    DataSetsLoader,
    RawDatabaseLoader,
    RawTableLoader,
    ResourceLoader,
    SpaceLoader,
    TransformationLoader,
    ViewLoader,
)
from cognite_toolkit._cdf_tk.utils import calculate_secure_hash
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables
from tests.test_unit.approval_client import ApprovalToolkitClient


def _return_none(*args, **kwargs) -> str | None:
    return None


class TestTransformationLoader:
    trafo_yaml = """
externalId: tr_first_transformation
name: 'example:first:transformation'
interval: '{{scheduleHourly}}'
isPaused: true
query: "INLINE"
destination:
  type: 'assets'
ignoreNullFields: true
isPublic: true
conflictMode: upsert
"""
    trafo_sql = "FILE"

    def test_no_auth_load(
        self,
        toolkit_client_approval: ApprovalToolkitClient,
        env_vars_with_client: EnvironmentVariables,
    ) -> None:
        loader = TransformationLoader(toolkit_client_approval.mock_client, None)
        filepath = self._create_mock_file(self.trafo_yaml)

        raw_list = loader.load_resource_file(filepath, env_vars_with_client.dump())
        loaded = loader.load_resource(raw_list[0], is_dry_run=False)

        assert loaded.destination_oidc_credentials is None
        assert loaded.source_oidc_credentials is None

    def test_oidc_auth_load(
        self,
        toolkit_client_approval: ApprovalToolkitClient,
        env_vars_with_client: EnvironmentVariables,
        monkeypatch: MonkeyPatch,
    ) -> None:
        loader = TransformationLoader(toolkit_client_approval.mock_client, None)
        resource = yaml.CSafeLoader(self.trafo_yaml).get_data()
        resource["authentication"] = {
            "clientId": "my-client-id",
            "clientSecret": "my-client-secret",
            "tokenUri": "https://cognite.com/token",
            "cdfProjectName": "my-project",
            "scopes": "USER_IMPERSONATION",
            "audience": "https://cognite.com",
        }
        filepath = self._create_mock_file(yaml.dump(resource))
        resource_id = resource["externalId"]

        raw_list = loader.load_resource_file(filepath, env_vars_with_client.dump())
        _ = loader.load_resource(raw_list[0], is_dry_run=False)

        read_credentials = loader._authentication_by_id_operation[(resource_id, "read")]
        write_credentials = loader._authentication_by_id_operation[(resource_id, "write")]

        assert read_credentials.dump() == write_credentials.dump()

    @staticmethod
    def _create_mock_file(content: str) -> MagicMock:
        filepath = MagicMock(spec=Path)
        filepath.read_text.return_value = content
        filepath.name = "transformation.yaml"
        filepath.stem = "transformation"
        filepath.parent = Path("path")
        return filepath

    def test_auth_unchanged_changed(
        self,
        toolkit_client_approval: ApprovalToolkitClient,
    ) -> None:
        local_content = """name: my-transformation
externalId: my_transformation
ignoreNullFields: true
query: SELECT * FROM my_table
authentication:
  clientId: my-client-id
  clientSecret: my-client-secret
  scopes: USER_IMPERSONATION
  tokenUri: https://cognite.com/token
  cdfProjectName: my-project
        """
        auth_dict = {"authentication": yaml.CSafeLoader(local_content).get_data()["authentication"]}
        auth_hash = calculate_secure_hash(auth_dict, shorten=True)
        cdf_transformation = Transformation(
            name="my-transformation",
            external_id="my_transformation",
            query=f"{TransformationLoader._hash_key}: {auth_hash}\nSELECT * FROM my_table",
            ignore_null_fields=True,
        )
        with monkeypatch_toolkit_client() as client:
            loader = TransformationLoader(client, None, None)

        filepath = self._create_mock_file(local_content)
        local_dumped = loader.load_resource_file(filepath, {})[0]
        cdf_dumped = loader.dump_resource(cdf_transformation, local_dumped)
        assert cdf_dumped == local_dumped

        new_filepath = self._create_mock_file(local_content.replace("my-client-secret", "my-new-client-secret"))
        new_local_dumped = loader.load_resource_file(new_filepath, {})[0]
        cdf_dumped = loader.dump_resource(cdf_transformation, new_local_dumped)
        assert cdf_dumped != new_local_dumped

    def test_sql_inline(
        self,
        toolkit_client_approval: ApprovalToolkitClient,
        env_vars_with_client: EnvironmentVariables,
        monkeypatch: MonkeyPatch,
    ) -> None:
        loader = TransformationLoader(toolkit_client_approval.mock_client, None)

        filepath = self._create_mock_file(self.trafo_yaml)
        resource = yaml.CSafeLoader(self.trafo_yaml).get_data()

        raw_list = loader.load_resource_file(filepath, env_vars_with_client.dump())
        loaded = loader.load_resource(raw_list[0], is_dry_run=False)
        assert loaded.query == resource["query"]

    @pytest.mark.parametrize(
        "item, expected",
        [
            pytest.param(
                {
                    "dataSetExternalId": "ds_my_dataset",
                    "destination": {
                        "type": "instances",
                        "dataModel": {
                            "space": "sp_model_space",
                            "externalId": "my_model",
                            "version": "v1",
                            "destinationType": "assets",
                        },
                        "instanceSpace": "sp_data_space",
                    },
                },
                [
                    (DataSetsLoader, "ds_my_dataset"),
                    (SpaceLoader, "sp_data_space"),
                    (DataModelLoader, dm.DataModelId(space="sp_model_space", external_id="my_model", version="v1")),
                ],
                id="Transformation to data model",
            ),
            pytest.param(
                {
                    "destination": {
                        "type": "nodes",
                        "view": {"space": "sp_space", "externalId": "my_view", "version": "v1"},
                        "instanceSpace": "sp_data_space",
                    }
                },
                [
                    (SpaceLoader, "sp_data_space"),
                    (ViewLoader, dm.ViewId(space="sp_space", external_id="my_view", version="v1")),
                ],
                id="Transformation to nodes ",
            ),
            pytest.param(
                {"destination": {"type": "raw", "database": "my_db", "table": "my_table"}},
                [
                    (RawDatabaseLoader, RawDatabase("my_db")),
                    (RawTableLoader, RawTable("my_db", "my_table")),
                ],
                id="Transformation to RAW table",
            ),
        ],
    )
    def test_get_dependent_items(self, item: dict, expected: list[tuple[type[ResourceLoader], Hashable]]) -> None:
        actual = TransformationLoader.get_dependent_items(item)

        assert list(actual) == expected



import time
from abc import ABC, abstractmethod
from collections.abc import Iterable, Iterator
from functools import lru_cache
from itertools import groupby
from pathlib import Path
from typing import Any, Callable, ClassVar, Generic, Literal

from cognite.client.data_classes import (
    Asset,
    AssetFilter,
    DataSetList,
    Event,
    EventFilter,
    FileMetadata,
    FileMetadataFilter,
    LabelDefinitionList,
    TimeSeries,
    TimeSeriesFilter,
)
from cognite.client.data_classes._base import T_CogniteResource
from rich.console import Console
from rich.progress import track

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.commands._base import ToolkitCommand
from cognite_toolkit._cdf_tk.exceptions import (
    ToolkitFileExistsError,
    ToolkitIsADirectoryError,
    ToolkitValueError,
)
from cognite_toolkit._cdf_tk.loaders import (
    AssetLoader,
    DataSetsLoader,
    EventLoader,
    FileMetadataLoader,
    LabelLoader,
    ResourceLoader,
    TimeSeriesLoader,
)
from cognite_toolkit._cdf_tk.utils import humanize_collection
from cognite_toolkit._cdf_tk.utils.cdf import metadata_key_counts
from cognite_toolkit._cdf_tk.utils.file import safe_rmtree
from cognite_toolkit._cdf_tk.utils.producer_worker import ProducerWorkerExecutor
from cognite_toolkit._cdf_tk.utils.table_writers import (
    FileFormat,
    Schema,
    SchemaColumn,
    SchemaColumnList,
    TableFileWriter,
)


class DataFinder:
    supported_formats: ClassVar[frozenset[FileFormat]] = frozenset()
    # This is the standard maximum items that can be returns by most CDF endpoints.
    chunk_size: ClassVar[int] = 1000

    def validate_format(self, format_: str) -> Literal[FileFormat]:
        if format_ in self.supported_formats:
            return format_  # type: ignore[return-value]
        raise ToolkitValueError(
            f"Unsupported format {format_}. Supported formats are {humanize_collection(self.supported_formats)}."
        )

    @abstractmethod
    def create_iterators(
        self, format_: FileFormat, limit: int | None
    ) -> Iterator[tuple[Schema, int, Iterable, Callable]]:
        """Create an iterator for the specified format."""
        raise NotImplementedError("This method should be implemented in subclasses.")


class AssetCentricFinder(DataFinder, ABC, Generic[T_CogniteResource]):
    def __init__(self, client: ToolkitClient, hierarchies: list[str], data_sets: list[str]):
        self.client = client
        self.hierarchies = hierarchies
        self.data_sets = data_sets
        self.loader = self._create_loader(client)
        self._hierarchy_set = set(self.hierarchies)
        self._data_set_set = set(self.data_sets)
        self._used_labels: set[str] = set()
        self._used_data_sets: set[str] = set()

    @abstractmethod
    def _create_loader(self, client: ToolkitClient) -> ResourceLoader:
        """Create the appropriate loader for the finder."""
        raise NotImplementedError()

    @lru_cache
    def aggregate_count(self, hierarchies: tuple[str, ...], data_sets: tuple[str, ...]) -> int:
        return self._aggregate_count(list(hierarchies), list(data_sets))

    @abstractmethod
    def _aggregate_count(self, hierarchies: list[str], data_sets: list[str]) -> int:
        raise NotImplementedError()

    @abstractmethod
    def _get_resource_columns(self) -> SchemaColumnList:
        """Get the columns for the schema."""
        raise NotImplementedError()

    @abstractmethod
    def create_resource_iterator(self, limit: int | None) -> Iterable:
        raise NotImplementedError()

    @abstractmethod
    def _resource_processor(self, items: Iterable[T_CogniteResource]) -> list[tuple[str, list[dict[str, Any]]]]:
        """Process the resources and return them in a format suitable for writing."""
        raise NotImplementedError()

    def _to_write(self, items: Iterable[T_CogniteResource]) -> list[dict[str, Any]]:
        write_items: list[dict[str, Any]] = []
        for item in items:
            dumped = self.loader.dump_resource(item)
            if "metadata" in dumped:
                metadata = dumped.pop("metadata")
                for key, value in metadata.items():
                    dumped[f"metadata.{key}"] = value
            if isinstance(dumped.get("labels"), list):
                dumped["labels"] = [label["externalId"] for label in dumped["labels"]]
                self._used_labels.update(dumped["labels"])
            if "dataSetExternalId" in dumped:
                self._used_data_sets.add(dumped["dataSetExternalId"])
            write_items.append(dumped)
        return write_items

    def create_iterators(
        self, format_: FileFormat, limit: int | None
    ) -> Iterator[tuple[Schema, int, Iterable, Callable]]:
        total = self.aggregate_count(tuple(self.hierarchies), tuple(self.data_sets))
        columns = self._get_resource_columns()

        iteration_count = total // self.chunk_size + (1 if total % self.chunk_size > 0 else 0)
        if iteration_count == 0:
            return

        yield (
            Schema(
                display_name=self.loader.display_name,
                format_=format_,
                columns=columns,
                folder_name=self.loader.folder_name,
                kind=self.loader.kind,
            ),
            iteration_count,
            self.create_resource_iterator(limit),
            self._resource_processor,
        )
        if self._used_data_sets:
            yield self._data_sets()
        if self._used_labels:
            yield self._labels()

    def _data_sets(self) -> tuple[Schema, int, Iterable, Callable]:
        data_sets = self.client.data_sets.retrieve_multiple(
            external_ids=list(self._used_data_sets), ignore_unknown_ids=True
        )
        loader = DataSetsLoader.create_loader(self.client)

        def process_data_sets(items: DataSetList) -> list[tuple[str, list[dict[str, Any]]]]:
            # All data sets are written to a single group, thus the empty string as the group key.
            # (Group keys are for example used in CSV files to create separate files for each
            # data set an asset belongs to.)
            return [("", [loader.dump_resource(item) for item in items])]

        return (
            # YAML format does not need columns.
            Schema(
                display_name=loader.display_name,
                format_="yaml",
                columns=SchemaColumnList(),
                folder_name=loader.folder_name,
                kind=loader.kind,
            ),
            1,
            [data_sets],
            process_data_sets,
        )

    def _labels(self) -> tuple[Schema, int, Iterable, Callable]:
        labels = self.client.labels.retrieve(external_id=list(self._used_labels))
        loader = LabelLoader.create_loader(self.client)

        def process_labels(items: LabelDefinitionList) -> list[tuple[str, list[dict[str, Any]]]]:
            # All labels are written to a single group, thus the empty string as the group key.
            # (Group keys are for example used in CSV files to create separate files for each
            # label an asset belongs to.)
            return [("", [loader.dump_resource(item) for item in items])]

        return (
            # YAML format does not need columns.
            Schema(
                display_name=loader.display_name,
                format_="yaml",
                columns=SchemaColumnList(),
                folder_name=loader.folder_name,
                kind=loader.kind,
            ),
            1,
            [labels],
            process_labels,
        )


class AssetFinder(AssetCentricFinder[Asset]):
    supported_formats = frozenset({"csv", "parquet", "yaml"})

    def _create_loader(self, client: ToolkitClient) -> ResourceLoader:
        return AssetLoader.create_loader(client)

    def _aggregate_count(self, hierarchies: list[str], data_sets: list[str]) -> int:
        return self.client.assets.aggregate_count(
            filter=AssetFilter(
                data_set_ids=[{"externalId": item} for item in data_sets] or None,
                asset_subtree_ids=[{"externalId": item} for item in hierarchies] or None,
            )
        )

    def create_resource_iterator(self, limit: int | None) -> Iterator:
        return self.client.assets(
            chunk_size=self.chunk_size,
            asset_subtree_external_ids=self.hierarchies or None,
            data_set_external_ids=self.data_sets or None,
            limit=limit,
        )

    def _resource_processor(self, assets: Iterable[Asset]) -> list[tuple[str, list[dict[str, Any]]]]:
        grouped_assets: list[tuple[str, list[dict[str, object]]]] = []
        for group, asset_group in groupby(
            sorted([(self._group(asset), asset) for asset in assets], key=lambda x: x[0]), key=lambda x: x[0]
        ):
            grouped_assets.append((group, self._to_write([asset for _, asset in asset_group])))
        return grouped_assets

    def _group(self, item: Asset) -> str:
        if self.hierarchies and self.data_sets:
            asset_external_id = self.client.lookup.assets.external_id(item.root_id or 0)
            data_set_external_id = self.client.lookup.data_sets.external_id(item.data_set_id or 0)
            if asset_external_id and data_set_external_id:
                return f"{asset_external_id}.{data_set_external_id}"
            elif asset_external_id:
                return asset_external_id
            elif data_set_external_id:
                return data_set_external_id
            return ""
        elif self.hierarchies:
            return self.client.lookup.assets.external_id(item.root_id or 0) or ""
        elif self.data_sets:
            return self.client.lookup.data_sets.external_id(item.data_set_id or 0) or ""
        return ""

    def _get_resource_columns(self) -> SchemaColumnList:
        columns = SchemaColumnList(
            [
                SchemaColumn(name="externalId", type="string"),
                SchemaColumn(name="name", type="string"),
                SchemaColumn(name="parentExternalId", type="string"),
                SchemaColumn(name="description", type="string"),
                SchemaColumn(name="dataSetExternalId", type="string"),
                SchemaColumn(name="source", type="string"),
                SchemaColumn(name="labels", type="string", is_array=True),
                SchemaColumn(name="geoLocation", type="json"),
            ]
        )
        data_set_ids = self.client.lookup.data_sets.id(self.data_sets) if self.data_sets else []
        root_ids = self.client.lookup.assets.id(self.hierarchies) if self.hierarchies else []
        metadata_keys = metadata_key_counts(self.client, "assets", data_set_ids or None, root_ids or None)
        sorted_keys = sorted([key for key, count in metadata_keys if count > 0])
        columns.extend([SchemaColumn(name=f"metadata.{key}", type="string") for key in sorted_keys])
        return columns


class FileMetadataFinder(AssetCentricFinder[FileMetadata]):
    supported_formats = frozenset({"csv", "parquet"})

    def _create_loader(self, client: ToolkitClient) -> ResourceLoader:
        return FileMetadataLoader.create_loader(client)

    def _aggregate_count(self, hierarchies: list[str], data_sets: list[str]) -> int:
        result = self.client.files.aggregate(
            filter=FileMetadataFilter(
                data_set_ids=[{"externalId": item} for item in data_sets] or None,
                asset_subtree_ids=[{"externalId": item} for item in hierarchies] or None,
            )
        )
        return result[0].count if result else 0

    def _get_resource_columns(self) -> SchemaColumnList:
        columns = SchemaColumnList(
            [
                SchemaColumn(name="externalId", type="string"),
                SchemaColumn(name="name", type="string"),
                SchemaColumn(name="directory", type="string"),
                SchemaColumn(name="source", type="string"),
                SchemaColumn(name="mimeType", type="string"),
                SchemaColumn(name="assetExternalIds", type="string", is_array=True),
                SchemaColumn(name="dataSetExternalId", type="string"),
                SchemaColumn(name="sourceCreatedTime", type="integer"),
                SchemaColumn(name="sourceModifiedTime", type="integer"),
                SchemaColumn(name="securityCategories", type="string", is_array=True),
                SchemaColumn(name="labels", type="string", is_array=True),
                SchemaColumn(name="geoLocation", type="json"),
            ]
        )
        data_set_ids = self.client.lookup.data_sets.id(self.data_sets) if self.data_sets else []
        root_ids = self.client.lookup.assets.id(self.hierarchies) if self.hierarchies else []
        metadata_keys = metadata_key_counts(self.client, "files", data_set_ids or None, root_ids or None)
        sorted_keys = sorted([key for key, count in metadata_keys if count > 0])
        columns.extend([SchemaColumn(name=f"metadata.{key}", type="string") for key in sorted_keys])
        return columns

    def create_resource_iterator(self, limit: int | None) -> Iterable:
        return self.client.files(
            chunk_size=self.chunk_size,
            asset_subtree_external_ids=self.hierarchies or None,
            data_set_external_ids=self.data_sets or None,
            limit=limit,
        )

    def _resource_processor(self, items: Iterable[FileMetadata]) -> list[tuple[str, list[dict[str, Any]]]]:
        return [("", self._to_write(items))]


class TimeSeriesFinder(AssetCentricFinder[TimeSeries]):
    supported_formats = frozenset({"csv", "parquet", "yaml"})

    def _create_loader(self, client: ToolkitClient) -> TimeSeriesLoader:
        return TimeSeriesLoader.create_loader(client)

    def _aggregate_count(self, hierarchies: list[str], data_sets: list[str]) -> int:
        return self.client.time_series.aggregate_count(
            filter=TimeSeriesFilter(
                data_set_ids=[{"externalId": item} for item in data_sets] or None,
                asset_subtree_ids=[{"externalId": item} for item in hierarchies] or None,
            )
        )

    def create_resource_iterator(self, limit: int | None) -> Iterator:
        return self.client.time_series(
            chunk_size=self.chunk_size,
            asset_subtree_external_ids=self.hierarchies or None,
            data_set_external_ids=self.data_sets or None,
            limit=limit,
        )

    def _resource_processor(self, time_series: Iterable[TimeSeries]) -> list[tuple[str, list[dict[str, Any]]]]:
        return [("", self._to_write(time_series))]

    def _get_resource_columns(self) -> SchemaColumnList:
        columns = SchemaColumnList(
            [
                SchemaColumn(name="externalId", type="string"),
                SchemaColumn(name="name", type="string"),
                SchemaColumn(name="isString", type="boolean"),
                SchemaColumn(name="unit", type="string"),
                SchemaColumn(name="unitExternalId", type="string"),
                SchemaColumn(name="assetExternalId", type="string"),
                SchemaColumn(name="isStep", type="boolean"),
                SchemaColumn(name="description", type="string"),
                SchemaColumn(name="dataSetExternalId", type="string"),
                SchemaColumn(name="securityCategories", type="string", is_array=True),
            ]
        )
        data_set_ids = self.client.lookup.data_sets.id(self.data_sets) if self.data_sets else []
        root_ids = self.client.lookup.assets.id(self.hierarchies) if self.hierarchies else []
        metadata_keys = metadata_key_counts(self.client, "timeseries", data_set_ids or None, root_ids or None)
        sorted_keys = sorted([key for key, count in metadata_keys if count > 0])
        columns.extend([SchemaColumn(name=f"metadata.{key}", type="string") for key in sorted_keys])
        return columns


class EventFinder(AssetCentricFinder[Event]):
    supported_formats = frozenset({"csv", "parquet"})

    def _create_loader(self, client: ToolkitClient) -> ResourceLoader:
        return EventLoader.create_loader(client)

    def _aggregate_count(self, hierarchies: list[str], data_sets: list[str]) -> int:
        return self.client.events.aggregate_count(
            filter=EventFilter(
                data_set_ids=[{"externalId": item} for item in data_sets] or None,
                asset_subtree_ids=[{"externalId": item} for item in hierarchies] or None,
            )
        )

    def _get_resource_columns(self) -> SchemaColumnList:
        columns = SchemaColumnList(
            [
                SchemaColumn(name="externalId", type="string"),
                SchemaColumn(name="dataSetExternalId", type="string"),
                SchemaColumn(name="startTime", type="integer"),
                SchemaColumn(name="endTime", type="integer"),
                SchemaColumn(name="type", type="string"),
                SchemaColumn(name="subtype", type="string"),
                SchemaColumn(name="description", type="string"),
                SchemaColumn(name="assetExternalIds", type="string", is_array=True),
                SchemaColumn(name="source", type="string"),
            ]
        )
        data_set_ids = self.client.lookup.data_sets.id(self.data_sets) if self.data_sets else []
        root_ids = self.client.lookup.assets.id(self.hierarchies) if self.hierarchies else []
        metadata_keys = metadata_key_counts(self.client, "events", data_set_ids or None, root_ids or None)
        sorted_keys = sorted([key for key, count in metadata_keys if count > 0])
        columns.extend([SchemaColumn(name=f"metadata.{key}", type="string") for key in sorted_keys])
        return columns

    def create_resource_iterator(self, limit: int | None) -> Iterable:
        return self.client.events(
            chunk_size=self.chunk_size,
            asset_subtree_external_ids=self.hierarchies or None,
            data_set_external_ids=self.data_sets or None,
            limit=limit,
        )

    def _resource_processor(self, items: Iterable[Event]) -> list[tuple[str, list[dict[str, Any]]]]:
        return [("", self._to_write(items))]


class DumpDataCommand(ToolkitCommand):
    def dump_table(
        self,
        finder: DataFinder,
        output_dir: Path,
        clean: bool,
        limit: int | None = None,
        format_: str = "csv",
        verbose: bool = False,
        parallel_threshold: int = 10,
        max_queue_size: int = 10,
    ) -> None:
        """Dumps data from CDF to a file

        Args:
            finder (DataFinder): The finder object to use for fetching data.
            output_dir (Path): The directory to write the output files to.
            clean (bool): Whether to clean the output directory before writing files.
            limit (int | None, optional): The maximum number of rows to write. Defaults to None.
            format_ (Literal["yaml", "csv", "parquet"], optional): The format of the output file. Defaults to "csv".
            verbose (bool, optional): Whether to print detailed progress information. Defaults to False.
            parallel_threshold (int, optional): The iteration threshold for parallel processing. Defaults to 10.
            max_queue_size (int, optional): If using parallel processing, the maximum size of the queue. Defaults to 10.

        """
        valid_format = finder.validate_format(format_)
        self.validate_directory(output_dir, clean)

        console = Console()
        # The ignore is used as MyPy does not understand that is_supported_format
        # above guarantees that the format is valid.
        for schema, iteration_count, resource_iterator, resource_processor in finder.create_iterators(
            valid_format, limit
        ):
            writer_cls = TableFileWriter.get_write_cls(schema.format_)
            row_counts = 0
            t0 = time.perf_counter()
            with writer_cls(schema, output_dir) as writer:
                if iteration_count > parallel_threshold:
                    executor = ProducerWorkerExecutor(
                        download_iterable=resource_iterator,
                        process=resource_processor,
                        write=writer.write_rows,
                        iteration_count=iteration_count,
                        max_queue_size=max_queue_size,
                    )
                    executor.run()
                    if executor.error_occurred:
                        raise ToolkitValueError(executor.error_message)
                    row_counts = executor.total_items
                else:
                    for resources in track(
                        resource_iterator, total=iteration_count, description=f"Dumping {schema.display_name}"
                    ):
                        row_counts += len(resources)
                        processed = resource_processor(resources)
                        writer.write_rows(processed)
            elapsed = time.perf_counter() - t0
            console.print(f"Dumped {row_counts:,} rows to {output_dir} in {elapsed:,.2f} seconds.")

    @staticmethod
    def validate_directory(output_dir: Path, clean: bool) -> None:
        if output_dir.exists() and clean:
            safe_rmtree(output_dir)
        elif output_dir.exists():
            raise ToolkitFileExistsError(f"Output directory {output_dir!s} already exists. Use --clean to remove it.")
        elif output_dir.suffix:
            raise ToolkitIsADirectoryError(f"Output directory {output_dir!s} is not a directory.")



from pathlib import Path
from typing import Annotated, Any, Optional, Union

import typer
from cognite.client.data_classes import WorkflowVersionId
from cognite.client.data_classes.data_modeling import DataModelId, ViewId
from rich import print

from cognite_toolkit._cdf_tk.commands import DumpDataCommand, DumpResourceCommand
from cognite_toolkit._cdf_tk.commands.dump_data import (
    AssetFinder,
    EventFinder,
    FileMetadataFinder,
    TimeSeriesFinder,
)
from cognite_toolkit._cdf_tk.commands.dump_resource import (
    DataModelFinder,
    GroupFinder,
    LocationFilterFinder,
    NodeFinder,
    TransformationFinder,
    WorkflowFinder,
)
from cognite_toolkit._cdf_tk.exceptions import ToolkitRequiredValueError
from cognite_toolkit._cdf_tk.feature_flags import Flags
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables
from cognite_toolkit._cdf_tk.utils.interactive_select import (
    AssetInteractiveSelect,
    EventInteractiveSelect,
    FileMetadataInteractiveSelect,
    TimeSeriesInteractiveSelect,
)


class DumpApp(typer.Typer):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.callback(invoke_without_command=True)(self.dump_main)
        if Flags.DUMP_DATA.is_enabled():
            self.add_typer(DumpDataApp(*args, **kwargs), name="data")
            self.add_typer(DumpConfigApp(*args, **kwargs), name="config")
        else:
            self.command("datamodel")(DumpConfigApp.dump_datamodel_cmd)

            self.command("asset")(DumpDataApp.dump_asset_cmd)
            self.command("timeseries")(DumpDataApp.dump_timeseries_cmd)

            self.command("workflow")(DumpConfigApp.dump_workflow)
            self.command("transformation")(DumpConfigApp.dump_transformation)
            self.command("group")(DumpConfigApp.dump_group)
            self.command("node")(DumpConfigApp.dump_node)

    @staticmethod
    def dump_main(ctx: typer.Context) -> None:
        """Commands to dump resource configurations from CDF into a temporary directory."""
        if ctx.invoked_subcommand is None:
            print("Use [bold yellow]cdf dump --help[/] for more information.")
        return None


class DumpConfigApp(typer.Typer):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.callback(invoke_without_command=True)(self.dump_config_main)

        self.command("datamodel")(self.dump_datamodel_cmd)
        self.command("workflow")(self.dump_workflow)
        self.command("transformation")(self.dump_transformation)
        self.command("group")(self.dump_group)
        self.command("node")(self.dump_node)
        if Flags.DUMP_EXTENDED.is_enabled():
            self.command("location-filters")(self.dump_location_filters)

    @staticmethod
    def dump_config_main(ctx: typer.Context) -> None:
        """Commands to dump resource configurations from CDF into a temporary directory."""
        if ctx.invoked_subcommand is None:
            print("Use [bold yellow]cdf dump config --help[/] for more information.")
        return None

    @staticmethod
    def dump_datamodel_cmd(
        ctx: typer.Context,
        data_model_id: Annotated[
            Optional[list[str]],
            typer.Argument(
                help="Data model ID to dump. Format: space external_id version. Example: 'my_space my_external_id v1'. "
                "Note that version is optional and defaults to the latest published version. If nothing is provided,"
                "an interactive prompt will be shown to select the data model.",
            ),
        ] = None,
        output_dir: Annotated[
            Path,
            typer.Option(
                "--output-dir",
                "-o",
                help="Where to dump the datamodel files.",
                allow_dash=True,
            ),
        ] = Path("tmp"),
        include_global: Annotated[
            bool,
            typer.Option(
                "--include-global",
                "-i",
                help="Include global containers, views, spaces in the dump. "
                "If this flag is not set, the global resources will be skipped.",
            ),
        ] = False,
        clean: Annotated[
            bool,
            typer.Option(
                "--clean",
                "-c",
                help="Delete the output directory before dumping the datamodel.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """This command will dump the selected data model as yaml to the folder specified, defaults to /tmp."""
        selected_data_model: Union[DataModelId, None] = None
        if data_model_id is not None:
            if len(data_model_id) < 2:
                raise ToolkitRequiredValueError(
                    "Data model ID must have at least 2 parts: space, external_id, and, optionally, version."
                )
            selected_data_model = DataModelId(*data_model_id)
        client = EnvironmentVariables.create_from_environment().get_client()

        cmd = DumpResourceCommand()
        cmd.run(
            lambda: cmd.dump_to_yamls(
                DataModelFinder(client, selected_data_model, include_global=include_global),
                output_dir=output_dir,
                clean=clean,
                verbose=verbose,
            )
        )

    @staticmethod
    def dump_workflow(
        ctx: typer.Context,
        workflow_id: Annotated[
            Optional[list[str]],
            typer.Argument(
                help="Workflow ID to dump. Format: external_id version. Example: 'my_external_id v1'. "
                "If nothing is provided, an interactive prompt will be shown to select the workflow",
            ),
        ] = None,
        output_dir: Annotated[
            Path,
            typer.Option(
                "--output-dir",
                "-o",
                help="Where to dump the workflow files.",
                allow_dash=True,
            ),
        ] = Path("tmp"),
        clean: Annotated[
            bool,
            typer.Option(
                "--clean",
                "-c",
                help="Delete the output directory before dumping the workflow.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """This command will dump the selected workflow as yaml to the folder specified, defaults to /tmp."""
        selected_workflow: Union[WorkflowVersionId, None] = None
        if workflow_id is not None:
            if len(workflow_id) <= 1:
                raise ToolkitRequiredValueError(
                    "Workflow ID must have at least 1 part: external_id and, optionally, version."
                )
            selected_workflow = WorkflowVersionId(*workflow_id)
        client = EnvironmentVariables.create_from_environment().get_client()

        cmd = DumpResourceCommand()
        cmd.run(
            lambda: cmd.dump_to_yamls(
                WorkflowFinder(client, selected_workflow),
                output_dir=output_dir,
                clean=clean,
                verbose=verbose,
            )
        )

    @staticmethod
    def dump_transformation(
        ctx: typer.Context,
        transformation_id: Annotated[
            Optional[list[str]],
            typer.Argument(
                help="Transformation IDs to dump. Format: external_id. Example: 'my_external_id'. "
                "If nothing is provided, an interactive prompt will be shown to select the transformation(s).",
            ),
        ] = None,
        output_dir: Annotated[
            Path,
            typer.Option(
                "--output-dir",
                "-o",
                help="Where to dump the transformation files.",
                allow_dash=True,
            ),
        ] = Path("tmp"),
        clean: Annotated[
            bool,
            typer.Option(
                "--clean",
                "-c",
                help="Delete the output directory before dumping the transformation.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """This command will dump the selected transformation as yaml to the folder specified, defaults to /tmp."""
        client = EnvironmentVariables.create_from_environment().get_client()

        cmd = DumpResourceCommand()
        cmd.run(
            lambda: cmd.dump_to_yamls(
                TransformationFinder(client, tuple(transformation_id) if transformation_id else None),
                output_dir=output_dir,
                clean=clean,
                verbose=verbose,
            )
        )

    @staticmethod
    def dump_group(
        ctx: typer.Context,
        group_name: Annotated[
            Optional[str],
            typer.Argument(
                help="Group name to dump. Format: name. Example: 'my_group'. "
                "If nothing is provided, an interactive prompt will be shown to select the group.",
            ),
        ] = None,
        output_dir: Annotated[
            Path,
            typer.Option(
                "--output-dir",
                "-o",
                help="Where to dump the group files.",
                allow_dash=True,
            ),
        ] = Path("tmp"),
        clean: Annotated[
            bool,
            typer.Option(
                "--clean",
                "-c",
                help="Delete the output directory before dumping the group.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """This command will dump the selected group as yaml to the folder specified, defaults to /tmp."""
        client = EnvironmentVariables.create_from_environment().get_client()

        cmd = DumpResourceCommand()
        cmd.run(
            lambda: cmd.dump_to_yamls(
                GroupFinder(client, group_name),
                output_dir=output_dir,
                clean=clean,
                verbose=verbose,
            )
        )

    @staticmethod
    def dump_node(
        ctx: typer.Context,
        view_id: Annotated[
            Optional[list[str]],
            typer.Argument(
                help="The view with the node properties you want to dump. Format: space externalId version. Example: 'my_space my_external_id version'. "
                "If nothing is provided, an interactive prompt will be shown to select the view.",
            ),
        ] = None,
        output_dir: Annotated[
            Path,
            typer.Option(
                "--output-dir",
                "-o",
                help="Where to dump the node files.",
                allow_dash=True,
            ),
        ] = Path("tmp"),
        clean: Annotated[
            bool,
            typer.Option(
                "--clean",
                "-c",
                help="Delete the output directory before dumping the node.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """This command will dump the selected node as yaml to the folder specified, defaults to /tmp.
        The intended use case is to dump nodes which are used as configuration. It is not intended to dump
        large amounts of data.
        """
        client = EnvironmentVariables.create_from_environment().get_client()
        selected_view_id: Union[None, ViewId] = None
        if view_id is not None:
            if len(view_id) <= 2:
                raise ToolkitRequiredValueError(
                    "View ID must have at least 2 parts: space, external_id and, optionally, version."
                )
            selected_view_id = ViewId(*view_id)

        cmd = DumpResourceCommand()
        cmd.run(
            lambda: cmd.dump_to_yamls(
                NodeFinder(client, selected_view_id),
                output_dir=output_dir,
                clean=clean,
                verbose=verbose,
            )
        )

    @staticmethod
    def dump_location_filters(
        ctx: typer.Context,
        external_id: Annotated[
            Optional[list[str]],
            typer.Argument(
                help="The external IDs of the location filters you want to dump. You can provide multiple external IDs separated by spaces. "
                "If nothing is provided, an interactive prompt will be shown to select the location filters.",
            ),
        ] = None,
        output_dir: Annotated[
            Path,
            typer.Option(
                "--output-dir",
                "-o",
                help="Where to dump the location filters.",
                allow_dash=True,
            ),
        ] = Path("tmp"),
        clean: Annotated[
            bool,
            typer.Option(
                "--clean",
                "-c",
                help="Delete the output directory before dumping the location filters.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """This command will dump the selected location filters as yaml to the folder specified, defaults to /tmp."""
        client = EnvironmentVariables.create_from_environment().get_client()
        cmd = DumpResourceCommand()
        cmd.run(
            lambda: cmd.dump_to_yamls(
                LocationFilterFinder(client, tuple(external_id) if external_id else None),
                output_dir=output_dir,
                clean=clean,
                verbose=verbose,
            )
        )


class DumpDataApp(typer.Typer):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.callback(invoke_without_command=True)(self.dump_data_main)
        self.command("asset")(self.dump_asset_cmd)
        self.command("files-metadata")(self.dump_files_cmd)
        self.command("timeseries")(self.dump_timeseries_cmd)
        self.command("event")(self.dump_event_cmd)

    @staticmethod
    def dump_data_main(ctx: typer.Context) -> None:
        """Commands to dump data from CDF into a temporary directory."""
        if ctx.invoked_subcommand is None:
            print("Use [bold yellow]cdf dump data --help[/] for more information.")
        return None

    @staticmethod
    def dump_asset_cmd(
        ctx: typer.Context,
        hierarchy: Annotated[
            Optional[list[str]],
            typer.Option(
                "--hierarchy",
                "-h",
                help="Hierarchy to dump.",
            ),
        ] = None,
        data_set: Annotated[
            Optional[list[str]],
            typer.Option(
                "--data-set",
                "-d",
                help="Data set to dump. If neither hierarchy nor data set is provided, the user will be prompted"
                "to select which assets to dump",
            ),
        ] = None,
        format_: Annotated[
            str,
            typer.Option(
                "--format",
                "-f",
                help="Format to dump the assets in. Supported formats: csv, and parquet.",
            ),
        ] = "csv",
        limit: Annotated[
            Optional[int],
            typer.Option(
                "--limit",
                "-l",
                help="Limit the number of assets to dump.",
            ),
        ] = None,
        output_dir: Annotated[
            Path,
            typer.Option(
                "--output-dir",
                "-o",
                help="Where to dump the asset files.",
                allow_dash=True,
            ),
        ] = Path("tmp"),
        clean: Annotated[
            bool,
            typer.Option(
                "--clean",
                "-c",
                help="Delete the output directory before dumping the asset.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """This command will dump the selected assets in the selected format to the folder specified, defaults to /tmp."""
        cmd = DumpDataCommand()
        client = EnvironmentVariables.create_from_environment().get_client()
        if hierarchy is None and data_set is None:
            hierarchy, data_set = AssetInteractiveSelect(client).interactive_select_hierarchy_datasets()

        cmd.run(
            lambda: cmd.dump_table(
                AssetFinder(client, hierarchy or [], data_set or []),
                output_dir,
                clean,
                limit,
                format_,  # type: ignore [arg-type]
                verbose,
            )
        )

    @staticmethod
    def dump_files_cmd(
        ctx: typer.Context,
        hierarchy: Annotated[
            Optional[list[str]],
            typer.Option(
                "--hierarchy",
                "-h",
                help="Asset hierarchy (sub-trees) to dump filemetadata from.",
            ),
        ] = None,
        data_set: Annotated[
            Optional[list[str]],
            typer.Option(
                "--data-set",
                "-d",
                help="Data set to dump. If neither hierarchy nor data set is provided, the user will be prompted.",
            ),
        ] = None,
        format_: Annotated[
            str,
            typer.Option(
                "--format",
                "-f",
                help="Format to dump the filemetadata in. Supported formats: csv, and parquet.",
            ),
        ] = "csv",
        limit: Annotated[
            Optional[int],
            typer.Option(
                "--limit",
                "-l",
                help="Limit the number of filemetadata to dump.",
            ),
        ] = None,
        output_dir: Annotated[
            Path,
            typer.Option(
                "--output-dir",
                "-o",
                help="Where to dump the filemetadata files.",
                allow_dash=True,
            ),
        ] = Path("tmp"),
        clean: Annotated[
            bool,
            typer.Option(
                "--clean",
                "-c",
                help="Delete the output directory before dumping the filemetadata.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """This command will dump the selected events to the selected format in the folder specified, defaults to /tmp."""
        cmd = DumpDataCommand()
        cmd.validate_directory(output_dir, clean)
        client = EnvironmentVariables.create_from_environment().get_client()
        if hierarchy is None and data_set is None:
            hierarchy, data_set = FileMetadataInteractiveSelect(client).interactive_select_hierarchy_datasets()
        cmd.run(
            lambda: cmd.dump_table(
                FileMetadataFinder(client, hierarchy or [], data_set or []),
                output_dir,
                clean,
                limit,
                format_,  # type: ignore [arg-type]
                verbose,
            )
        )

    @staticmethod
    def dump_timeseries_cmd(
        ctx: typer.Context,
        hierarchy: Annotated[
            Optional[list[str]],
            typer.Option(
                "--hierarchy",
                "-h",
                help="Asset hierarchy (sub-trees) to dump timeseries from.",
            ),
        ] = None,
        data_set: Annotated[
            Optional[list[str]],
            typer.Option(
                "--data-set",
                "-d",
                help="Data set to dump. If neither hierarchy nor data set is provided, the user will be prompted.",
            ),
        ] = None,
        format_: Annotated[
            str,
            typer.Option(
                "--format",
                "-f",
                help="Format to dump the timeseries in. Supported formats: csv, and parquet.",
            ),
        ] = "csv",
        limit: Annotated[
            Optional[int],
            typer.Option(
                "--limit",
                "-l",
                help="Limit the number of timeseries to dump.",
            ),
        ] = None,
        output_dir: Annotated[
            Path,
            typer.Option(
                "--output-dir",
                "-o",
                help="Where to dump the timeseries files.",
                allow_dash=True,
            ),
        ] = Path("tmp"),
        clean: Annotated[
            bool,
            typer.Option(
                "--clean",
                "-c",
                help="Delete the output directory before dumping the timeseries.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """This command will dump the selected timeseries to the selected format in the folder specified, defaults to /tmp."""
        cmd = DumpDataCommand()
        client = EnvironmentVariables.create_from_environment().get_client()
        if hierarchy is None and data_set is None:
            hierarchy, data_set = TimeSeriesInteractiveSelect(client).interactive_select_hierarchy_datasets()
        cmd.run(
            lambda: cmd.dump_table(
                TimeSeriesFinder(client, hierarchy or [], data_set or []),
                output_dir,
                clean,
                limit,
                format_,  # type: ignore [arg-type]
                verbose,
            )
        )

    @staticmethod
    def dump_event_cmd(
        ctx: typer.Context,
        hierarchy: Annotated[
            Optional[list[str]],
            typer.Option(
                "--hierarchy",
                "-h",
                help="Asset hierarchy (sub-trees) to dump event from.",
            ),
        ] = None,
        data_set: Annotated[
            Optional[list[str]],
            typer.Option(
                "--data-set",
                "-d",
                help="Data set to dump. If neither hierarchy nor data set is provided, the user will be prompted.",
            ),
        ] = None,
        format_: Annotated[
            str,
            typer.Option(
                "--format",
                "-f",
                help="Format to dump the event in. Supported formats: csv, and parquet.",
            ),
        ] = "csv",
        limit: Annotated[
            Optional[int],
            typer.Option(
                "--limit",
                "-l",
                help="Limit the number of events to dump.",
            ),
        ] = None,
        output_dir: Annotated[
            Path,
            typer.Option(
                "--output-dir",
                "-o",
                help="Where to dump the events files.",
                allow_dash=True,
            ),
        ] = Path("tmp"),
        clean: Annotated[
            bool,
            typer.Option(
                "--clean",
                "-c",
                help="Delete the output directory before dumping the events.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """This command will dump the selected events to the selected format in the folder specified, defaults to /tmp."""
        cmd = DumpDataCommand()
        cmd.validate_directory(output_dir, clean)
        client = EnvironmentVariables.create_from_environment().get_client()
        if hierarchy is None and data_set is None:
            hierarchy, data_set = EventInteractiveSelect(client).interactive_select_hierarchy_datasets()
        cmd.run(
            lambda: cmd.dump_table(
                EventFinder(client, hierarchy or [], data_set or []),
                output_dir,
                clean,
                limit,
                format_,
                verbose,
            )
        )



from collections.abc import Iterable

import pytest

from cognite_toolkit._cdf_tk.utils.tarjan import tarjan


def tarjan_test_cases() -> Iterable:
    yield pytest.param(
        {},
        [],
        id="Empty graph",
    )
    yield pytest.param(
        {"A": set()},
        [{"A"}],
        id="Single node",
    )
    yield pytest.param(
        {"A": {"B"}, "B": set()},
        [{"B"}, {"A"}],
        id="Two nodes",
    )
    yield pytest.param(
        {"A": {"B"}, "B": {"A"}},
        [{"A", "B"}],
        id="Circular dependency",
    )
    yield pytest.param(
        {"A": {"B"}, "B": {"C"}, "C": {"A"}},
        [{"C", "B", "A"}],
        id="Circular dependency with three nodes",
    )


class TestTarjan:
    @pytest.mark.parametrize("graph, expected", tarjan_test_cases())
    def test_tarjan(self, graph: dict[str, set[str]], expected: list[set[str]]):
        result = tarjan(graph)
        assert result == expected



from collections.abc import Iterable
from pathlib import Path

import pytest

from cognite_toolkit._cdf_tk.resource_classes import GroupYAML
from cognite_toolkit._cdf_tk.tk_warnings.fileread import ResourceFormatWarning
from cognite_toolkit._cdf_tk.utils import read_yaml_content
from cognite_toolkit._cdf_tk.validation import validate_resource_yaml_pydantic
from tests.test_unit.utils import find_resources


def invalid_group_test_cases() -> Iterable:
    yield pytest.param(
        """name: gp_table_scoped_complete_org
sourceId: '1234567890123456789'
metadata:
  origin: cognite-toolkit
  governed: true
  groupNo: 0123
  maxUsers: .inf
  description: foo:bar
capabilities:
- rawAcl:
    actions:
    - READ
    - WRITE
    scope:
      tableScope:
        dbsToTables:
          db_complete_org: {}
""",
        {
            "In metadata the key 'groupNo' should be a valid string. Got 83 of type int. Hint: Use double quotes to force string.",
            "In metadata the key 'governed' should be a valid string. Got True of type bool. Hint: Use double quotes to force string.",
            "In metadata the key 'maxUsers' should be a valid string. Got inf of type float. Hint: Use double quotes to force string.",
            "In capabilities[1].scope.dbsToTables input should be a valid list. Got {}.",
        },
        id="Naughty metadata and non-standard rawAcl tableScope",
    )

    yield pytest.param(
        """- name: group1
  sourceId: '1234567890123456789'
  capabilities:
  - labelsAcl:
     actions:
     - READ
     scope:
       all: {}
- name: group2
  sourceId: '1234567890123456789'
  capabilities:
  - labelsAcl:
     actions:
     - WRITE-KING
     scope:
       all: {}
""",
        {"In item [2].capabilities[1].actions input should be 'READ' or 'WRITE'. Got 'WRITE-KING'."},
        id="Error in second group",
    )


class TestTimeSeriesTK:
    @pytest.mark.parametrize("data", list(find_resources("Group")))
    def test_load_valid_timeseries(self, data: dict[str, object]) -> None:
        loaded = GroupYAML.model_validate(data)

        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize("content, expected_errors", list(invalid_group_test_cases()))
    def test_invalid_group_error_messages(self, content: str, expected_errors: set[str]) -> None:
        """Test the validate_resource_yaml function for GroupYAML."""
        data = read_yaml_content(content)

        warning_list = validate_resource_yaml_pydantic(data, GroupYAML, Path("some_file.yaml"))
        assert len(warning_list) == 1
        format_warning = warning_list[0]
        assert isinstance(format_warning, ResourceFormatWarning)

        assert set(format_warning.errors) == expected_errors



from pathlib import Path
from unittest.mock import MagicMock

import pytest

from cognite_toolkit._cdf_tk.loaders import ResourceWorker, ViewLoader
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables


class TestDeployCommand:
    def test_load_files(self, env_vars_with_client: EnvironmentVariables) -> None:
        path = MagicMock(spec=Path)
        path.name = "my.View.yaml"
        path.read_text.return_value = VIEW_SOURCE_NONE
        worker = ResourceWorker(ViewLoader.create_loader(env_vars_with_client.get_client()))

        with pytest.raises(TypeError) as e:
            worker.load_resources([path], environment_variables={}, is_dry_run=True, verbose=False)

        assert e.value


VIEW_SOURCE_NONE = """- space: dm_domain_generic
  externalId: Equipment
  name: Equipment
  version: v1
  properties:
    id:
      container:
        space: dm_domain_generic
        externalId: Tag
      containerPropertyIdentifier: id
      type:
        list: false
        collation: ucs_basic
        type: text
      source:
"""
